unit UMain;

{
  Shell Pack
  Copyright (c) 1996-2001 Arisesoft
  For conditions of distribution and use, see LICENSE.

  File Manager Demo
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Menus, ActnList, StdCtrls, ExtCtrls, ToolWin, ImgList,
  IniFiles, AS_ShellCtrls;

const
  WM_ENTERSHELLLIST = WM_USER + 1;

type
  TMainForm = class(TForm)
    PopupMenu2: TPopupMenu;
    ActionList1: TActionList;
    ActionRefresh: TAction;
    ActionSelectAll: TAction;
    ActionInvertSelection: TAction;
    N3: TMenuItem;
    SelectAll2: TMenuItem;
    ActionNewSubFolder: TAction;
    N6: TMenuItem;
    ActionProperties: TAction;
    Filter2: TMenuItem;
    N10: TMenuItem;
    NewSubFolder2: TMenuItem;
    ToolBar1: TToolBar;
    Panel1: TPanel;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    Panel2: TPanel;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ActionPathAsRoot: TAction;
    PopupMenu4: TPopupMenu;
    WhatsThis4: TMenuItem;
    ActionHistoryBack: TAction;
    Panel4: TPanel;
    Edit1: TMemo;
    ImageList1: TImageList;
    Panel5: TPanel;
    ShellTree1: TShellTree;
    ShellList1: TShellList;
    Paste2: TMenuItem;
    InvertSelection2: TMenuItem;
    N8: TMenuItem;
    FolderProperties2: TMenuItem;
    N2: TMenuItem;
    ActionFolderProperties: TAction;
    ActionPaste: TAction;
    Splitter1: TSplitter;
    ImageList2: TImageList;
    Filter21: TMenuItem;
    Filter22: TMenuItem;
    Clear2: TMenuItem;
    PathAsRoot2: TMenuItem;
    StatusBar1: TStatusBar;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ShellList1Enter(Sender: TObject);
    procedure ActionSelectAllExecute(Sender: TObject);
    procedure ActionInvertSelectionExecute(Sender: TObject);
    procedure ActionPropertiesExecute(Sender: TObject);
    procedure ActionNewSubFolderExecute(Sender: TObject);
    procedure ActionNewSubFolderUpdate(Sender: TObject);
    procedure ActionSelectAllUpdate(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
    procedure ToolButton2Click(Sender: TObject);
    procedure ToolButton3Click(Sender: TObject);
    procedure ToolButton4Click(Sender: TObject);
    procedure ActionPathAsRootExecute(Sender: TObject);
    procedure ActionPathAsRootUpdate(Sender: TObject);
    procedure ActionHistoryBackExecute(Sender: TObject);
    procedure ActionHistoryBackUpdate(Sender: TObject);
    procedure ActionRefreshExecute(Sender: TObject);
    procedure ActionRefreshUpdate(Sender: TObject);
    procedure ActionPropertiesUpdate(Sender: TObject);
    procedure ActionInvertSelectionUpdate(Sender: TObject);
    procedure ActionFolderPropertiesExecute(Sender: TObject);
    procedure ActionFolderPropertiesUpdate(Sender: TObject);
    procedure PopupMenu2Popup(Sender: TObject);
    procedure ActionPasteExecute(Sender: TObject);
    procedure ShellTree1History(Sender: TObject);
    procedure ActionPasteUpdate(Sender: TObject);
    procedure Edit1Enter(Sender: TObject);
    procedure Edit1Exit(Sender: TObject);
    procedure Edit1KeyPress(Sender: TObject; var Key: Char);
    procedure Edit1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Clear2Click(Sender: TObject);
    procedure Filter21Click(Sender: TObject);
    procedure Filter22Click(Sender: TObject);
    procedure FormShortCut(var Msg: TWMKey; var Handled: Boolean);
    procedure StatusBar1Hint(Sender: TObject);
  private
    FIniFile: TIniFile;
    FCurrentPath: string;
    procedure WMEnterShellList(var Message: TMessage); message WM_ENTERSHELLLIST;
    function IsEditing: Boolean;
    procedure LoadSettings;
    procedure SaveSettings;
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses Math, FileCtrl, AS_ShellUtils;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  DesktopFont := True;
  Edit1.WantTabs := True;
{ ^ using of VK_TAB for auto filling of path, therefore Edit1 is TMemo }
  EnableAutoComplete(Edit1.Handle, True, False);
  FIniFile := TIniFile.Create(ChangeFileExt(Application.ExeName, '.ini'));
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FIniFile.Free;
end;

procedure TMainForm.WMEnterShellList(var Message: TMessage);
begin
  with ShellList1 do
  begin
    if Focused and (GetCaptureControl <> ShellList1) and
      (Items.Count > 0) and (ItemFocused = nil) and (Selected = nil) then
      with Items[0] do
      begin
        Selected := True;
        Focused := True;
        MakeVisible(False);
      end;
  end;
end;

procedure TMainForm.ShellList1Enter(Sender: TObject);
begin
  PostMessage(Handle, WM_ENTERSHELLLIST, 0, 0);
end;

procedure TMainForm.ActionPropertiesExecute(Sender: TObject);
begin
  if (ActiveControl = ShellTree1) or (ShellList1.SelCount = 0) then
    ShellTree1.DoVerb('properties')
  else
    ShellList1.DoVerb('properties');
end;

procedure TMainForm.ActionPropertiesUpdate(Sender: TObject);
begin
  ActionProperties.Enabled := not IsEditing;
end;

procedure TMainForm.ActionRefreshExecute(Sender: TObject);
begin
  ShellTree1.Refresh(ShellTree1.Items.GetFirstNode, False);
end;

procedure TMainForm.ActionRefreshUpdate(Sender: TObject);
begin
  ActionRefresh.Enabled := not IsEditing;
end;

procedure TMainForm.ActionSelectAllExecute(Sender: TObject);
begin
  ShellList1.SelectAll;
end;

procedure TMainForm.ActionSelectAllUpdate(Sender: TObject);
begin
  ActionSelectAll.Enabled := (ActiveControl <> Edit1) and
    not IsEditing and (ShellList1.Items.Count > 0) and
    (ShellList1.SelCount <> ShellList1.Items.Count);
end;

procedure TMainForm.ActionInvertSelectionExecute(Sender: TObject);
var
  I: Integer;
begin
  for I := 0 to Pred(ShellList1.Items.Count) do
    with ShellList1.Items[I] do Selected := not Selected;
end;

procedure TMainForm.ActionInvertSelectionUpdate(Sender: TObject);
begin
  ActionInvertSelection.Enabled := (ActiveControl <> Edit1) and
    not IsEditing and (ShellList1.Items.Count > 0);
end;

procedure TMainForm.ActionPasteExecute(Sender: TObject);
begin
  ShellTree1.DoVerb('paste');
end;

procedure TMainForm.ActionPasteUpdate(Sender: TObject);
begin
  ActionPaste.Enabled := ShellTree1.CanPaste;
end;

procedure TMainForm.ActionNewSubFolderExecute(Sender: TObject);
begin
  ShellTree1.NewSubFolder(ShellTree1.Selected);
end;

procedure TMainForm.ActionNewSubFolderUpdate(Sender: TObject);
var
  Node: TTreeNode;
begin
  Node := ShellTree1.Selected;
  ActionNewSubFolder.Enabled := (Node <> nil) and IsFileSystemPIDL(Node.Data);
end;

procedure TMainForm.ToolButton1Click(Sender: TObject);
begin
  ShellTree1.Root := 'rsDesktop';
end;

procedure TMainForm.ToolButton2Click(Sender: TObject);
begin
  ShellTree1.Root := 'rsPersonal';
end;

procedure TMainForm.ToolButton3Click(Sender: TObject);
begin
  ShellTree1.Root := 'rsDrives';
end;

procedure TMainForm.ToolButton4Click(Sender: TObject);
begin
  ShellTree1.Root := 'rsNetwork';
end;

procedure TMainForm.ActionPathAsRootExecute(Sender: TObject);
begin
  ShellTree1.Root := ShellTree1.Path;
end;

procedure TMainForm.ActionPathAsRootUpdate(Sender: TObject);
var
  Node: TTreeNode;
begin
  Node := ShellTree1.Selected;
  ActionPathAsRoot.Enabled := (Node <> nil) and (Node.AbsoluteIndex > 0) and
    IsFileSystemPIDL(Node.Data);
end;

procedure TMainForm.ActionHistoryBackExecute(Sender: TObject);
begin
  ShellTree1.HistoryBack;
end;

procedure TMainForm.ActionHistoryBackUpdate(Sender: TObject);
begin
  ActionHistoryBack.Enabled := ShellTree1.IsHistory;
end;

function TMainForm.IsEditing: Boolean;
begin
  Result := ShellTree1.IsEditing or ShellList1.IsEditing;
end;

procedure TMainForm.PopupMenu2Popup(Sender: TObject);
begin
  if ShellTree1.IsEditing then
    ShellTree1.Selected.EndEdit(True)
  else
  if ShellList1.IsEditing then
    ShellList1.ItemFocused.CancelEdit;
end;

procedure TMainForm.ActionFolderPropertiesExecute(Sender: TObject);
begin
  ShellTree1.DoVerb('properties');
end;

procedure TMainForm.ActionFolderPropertiesUpdate(Sender: TObject);
var
  Node: TTreeNode;
begin
  Node := ShellTree1.Selected;
  ActionFolderProperties.Enabled := (Node <> nil) and
    not IsDesktopPIDL(Node.Data);
end;

procedure TMainForm.ShellTree1History(Sender: TObject);
begin
  FCurrentPath := ShellTree1.NameForAddressBar;
  Edit1Exit(nil);
end;

procedure TMainForm.Edit1Enter(Sender: TObject);
begin
  Edit1.Text := FCurrentPath;
  Edit1.SelectAll;
end;

procedure TMainForm.Edit1Exit(Sender: TObject);
begin
  Edit1.Text := MinimizeName(FCurrentPath, Canvas, Edit1.Width - 10);
end;

procedure TMainForm.Edit1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
    ShellTree1.Path := Edit1.Text;
  if (Key = Ord('A')) and ([ssCtrl] = Shift) then
  begin
    Edit1.SelectAll;
    Key := 0;
  end;
  if Key = VK_TAB then Key := 0;
end;

procedure TMainForm.Edit1KeyPress(Sender: TObject; var Key: Char);
begin
  if Key = ^A then Key := #0;
end;

const
  SRegRoot = 'Root';
  SRegSplitter = 'Splitter';
  SRegBookmark = 'Bookmark';
  SRegLeft = 'Left';
  SRegTop = 'Top';
  SRegRight = 'Right';
  SRegBottom = 'Bottom';

procedure TMainForm.LoadSettings;
var
  R: TRect;
begin
  ShellList1.LoadColumns(FIniFile, Name);
  with FIniFile do
  begin
    R := BoundsRect;
    R.Left := ReadInteger(Name, SRegLeft, R.Left);
    R.Top := ReadInteger(Name, SRegTop, R.Top);
    R.Right := ReadInteger(Name, SRegRight, R.Right);
    R.Bottom := ReadInteger(Name, SRegBottom, R.Bottom);
    BoundsRect := R;
    Panel1.Height := ReadInteger(Name, SRegSplitter, Panel1.Height);

    try
      ShellTree1.Root := ReadString(Name, SRegRoot, ShellTree1.Root);
    except
      Application.HandleException(Self);
    end;
    ShellTree1.Bookmark := ReadString(Name, SRegBookmark, '');
  end;
  ShellTree1.StartHistory;
  ShellTree1.ShellList := ShellList1;
  ShellList1.UpdateRoot;
end;

procedure TMainForm.SaveSettings;
var
  R: TRect;
begin
  with FIniFile do
  begin
    R := BoundsRect;
    WriteInteger(Name, SRegLeft, R.Left);
    WriteInteger(Name, SRegTop, R.Top);
    WriteInteger(Name, SRegRight, R.Right);
    WriteInteger(Name, SRegBottom, R.Bottom);

    WriteInteger(Name, SRegSplitter, Panel1.Height);
    WriteString(Name, SRegBookmark, ShellTree1.Bookmark);
    WriteString(Name, SRegRoot, ShellTree1.Root);
  end;
  ShellList1.SaveColumns(FIniFile, Name);
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  LoadSettings;
end;

procedure TMainForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  SaveSettings;
end;

procedure TMainForm.Clear2Click(Sender: TObject);
begin
  ShellList1.Filter := '';
  ShellList1.UpdateRoot;
  Filter21.Checked := False;
  Filter22.Checked := False;
end;

procedure TMainForm.Filter21Click(Sender: TObject);
begin
  ShellList1.Filter := 'html,htm,txt';
  ShellList1.UpdateRoot;
  Filter21.Checked := True;
  Filter22.Checked := False;
end;

procedure TMainForm.Filter22Click(Sender: TObject);
begin
  ShellList1.Filter := 'pas,dfm,exe';
  ShellList1.UpdateRoot;
  Filter21.Checked := False;
  Filter22.Checked := True;
end;

const
  scF5 = VK_F5;
  scCtrlA = scCtrl + Ord('A');
  scCtrlShiftA = scCtrlA + scShift;
  scAltEnter = scAlt + VK_RETURN;

procedure TMainForm.FormShortCut(var Msg: TWMKey; var Handled: Boolean);
var
  ShortCut: TShortCut;
  ShiftState: TShiftState;
begin
  ShiftState := KeyDataToShiftState(Msg.KeyData);
  ShortCut := Menus.ShortCut(Msg.CharCode, ShiftState);
  case ShortCut of
    scF5:
      Handled := ActionRefresh.Execute;
    scCtrlA:
      Handled := ActionSelectAll.Execute;
    scCtrlShiftA:
      Handled := ActionInvertSelection.Execute;
    scAltEnter:
      Handled := ActionProperties.Execute;
  else
    Handled := False;
  end;
end;

procedure TMainForm.StatusBar1Hint(Sender: TObject);
var
  S: string;
begin
  S := Application.Hint;
  if S <> '' then
  begin
    StatusBar1.Panels[0].Text := S;
    StatusBar1.Panels[0].Bevel := pbNone;
  end else
    StatusBar1.Panels[0].Bevel := pbLowered;
end;

end.

