unit AS_ShellCtrls;

{
  Arisesoft Shell Pack
  Copyright (c) 1996-2001 Arisesoft
  For conditions of distribution and use, see LICENSE.

  Shell Controls
}

{$I AS_Ver.inc}

interface

uses
  Windows, Messages, Classes, Controls, CommCtrl, ComCtrls, Graphics, Menus,
  ShlObj, ActiveX, IniFiles, Registry;

type
  TRoot = type string;

  TCustomShellList = class;

  TCustomShellTree = class(TCustomTreeView, IDropTarget)
  private
    FRoot,
      FOldRoot: TRoot;
    FUpdating: Boolean;
    FAutoContext: Boolean;
    FContextMenu2: IContextMenu2;
    FNotifyHandle: THandle;
    FDataObject: IDataObject;
    FDropLeftButton: Boolean;
    FLastAutoScrollTick: Cardinal;
    FStartAutoExpandTick: Cardinal;
    FDropMode: Boolean;
    FRegisterDragDrop: Boolean;
    FNextViewerHandle: THandle;
    FClipboardViewer: Boolean;
    FCutList: TList;
    FCutPIDL: PItemIDList;
    FHistoryList: TList;
    FLockHistory: Boolean;
    FOnHistory: TNotifyEvent;
    FOnRenamed: TTVChangedEvent;
    FShellList: TCustomShellList;
    function GetPath: string;
    function GetBookmark: string;
    procedure SetPath(const Value: string);
    procedure SetRoot(const Value: TRoot);
    procedure SetBookmark(const Value: string);
    procedure AddHistoryItem(PIDL: PItemIDList);
    procedure DoCompare(Sender: TObject; Node1, Node2: TTreeNode; Data: Integer; var Compare: Integer);
    procedure DoSelectFromPIDL(PIDL: PItemIDList);
    procedure DoRefresh(Node: TTreeNode; OnlyExpanded: Boolean);
    procedure CreateRoot;
    procedure NotifyMedia(PIDL: PItemIDList);
    procedure HandleNotify(WParam, LParam: Integer);
    procedure NotifyRename(OldPIDL, NewPIDL: PItemIDList);
    procedure DoNodeFromPIDL(PIDL: PItemIDList; out Node: TTreeNode; LastFound: Boolean); overload;
    procedure DoNodeFromPIDL(PIDL: PItemIDList; out Node, ParentNode: TTreeNode); overload;
    procedure NotifyDel(PIDL: PItemIDList);
    procedure DeleteFromParent(Node, ParentNode: TTreeNode);
    procedure DoBackRoot(Node: TTreeNode);
    procedure DoNotifyDel(Node, ParentNode: TTreeNode; out IsBackRoot: Boolean);
    procedure DoNotifyWin9xFix;
    procedure DoNotifyAdd(Node: TTreeNode; ParentNode: TTreeNode; PIDL: PItemIDList);
    procedure NotifyAdd(PIDL: PItemIDList);
    procedure NotifyShare(PIDL: PItemIDList);
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure NotifyUpdateItem(PIDL: PItemIDList);
    procedure DoDestroyWnd;
    procedure NotifyClipboard;
    function InCutList(PIDL: PItemIdList): Boolean;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure OleDragDrop(Node: TTreeNode; IsRightButton: Boolean);
    function GetDropTarget(MousePos: TPoint): IDropTarget;
    function GetNodeFromItem(const Item: TTVItem): TTreeNode;
    function GetContextMenu(out ContextMenu: IContextMenu; PIDL: PItemIDList): Boolean;
    procedure SetShellList(Value: TCustomShellList);
    procedure Editing(Node: TTreeNode);
    procedure UpdateShellList(Node: TTreeNode);
    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; stdcall;
    function IDropTarget_DragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;
    function IDropTarget.DragOver = IDropTarget_DragOver;
    function DragLeave: HResult; stdcall;
    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;
  protected
    procedure Delete(Node: TTreeNode); override;
    procedure DoAdded(Node: TTreeNode);
    procedure Change(Node: TTreeNode); override;
    function CanChange(Node: TTreeNode): Boolean; override;
    function CanExpand(Node: TTreeNode): Boolean; override;
    function CanEdit(Node: TTreeNode): Boolean; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure DoContextPopup(MousePos: TPoint; var Handled: Boolean); override;
    procedure DoHistory; dynamic;
    procedure DoRenamed(Node: TTreeNode); dynamic;
    procedure Edit(const Item: TTVItem); override;
    procedure Loaded; override;
    function PopulateNode(Node: TTreeNode): Boolean;
    procedure RootChanged;
    procedure WndProc(var Message: TMessage); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Refresh(Node: TTreeNode; OnlyExpanded: Boolean = False);
    procedure NewSubFolder(Node: TTreeNode);
    procedure DoVerb(Verb: PChar);
    procedure BackRoot(Node: TTreeNode);
    function IsHistory: Boolean;
    procedure HistoryBack;
    procedure StartHistory;
    procedure SelectFromPIDL(PIDL: PItemIDList);
    procedure Rename(Node: TTreeNode; const NewName: string);
    function CanPaste: Boolean;
    procedure FastSelect(Node: TTreeNode);
    function NameForAddressBar: string;
    property AutoContext: Boolean read FAutoContext write FAutoContext;
    property Bookmark: string read GetBookmark write SetBookmark;
    property Path: string read GetPath write SetPath;
    property Root: TRoot read FRoot write SetRoot;
    property ShellList: TCustomShellList read FShellList write SetShellList;
    property OnHistory: TNotifyEvent read FOnHistory write FOnHistory;
    property OnRenamed: TTVChangedEvent read FOnRenamed write FOnRenamed;
    property Items;
  end;

  TShellTree = class(TCustomShellTree)
  published
    property AutoContext default True;
    property ShellList;
    property OnHistory;
    property OnRenamed;
    property Align;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property Color;
    property Ctl3D;
    property Constraints;
    property Enabled;
    property Font;
    property HideSelection;
    property HotTrack;
    property Indent;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property RowSelect;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property TabOrder;
    property TabStop default True;
    property ToolTips;
    property Visible;
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnContextPopup;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnDblClick;
    property OnDeletion;
    property OnEditing;
    property OnEnter;
    property OnExit;
    property OnExpanding;
    property OnExpanded;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  TColumnKind = (ckName, ckSize, ckType, ckModified);

  TCustomShellList = class(TCustomListView, IDropTarget)
  private
    FUpdating: Boolean;
    FSelectedEdit: Boolean;
    FRootPIDL: PItemIDList;
    FShellFolder: IShellFolder;
    FShellList: TList;
    FSortDescent: Boolean;
    FSortColumn: TColumnKind;
    FSortArrows: array[0..1] of TBitmap;
    FHeaderInstance: Pointer;
    FDefHeaderProc: Pointer;
    FHeaderHandle: HWND;
    FAutoContext: Boolean;
    FContextMenu2: IContextMenu2;
    FColumnsContext: Boolean;
    FShellTree: TCustomShellTree;
    FFileNotify: TThread;
    FColumnsMenu: TPopupMenu;
    FColumnsWidth: array[TColumnKind] of Integer;
    FDataObject: IDataObject;
    FDropLeftButton: Boolean;
    FRegisterDragDrop: Boolean;
    FNextViewerHandle: THandle;
    FClipboardViewer: Boolean;
    FCutList: TList;
    FCutPIDL: PItemIDList;
    FSelList: TList;
    FFocusedPIDL: PItemIDList;
    FFilter: TStringList;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure ClearShellList(IsRefresh: Boolean = False);
    procedure Sort;
    procedure SetSortColumn(Value: TColumnKind);
    procedure ClearHeader;
    procedure UpdateHeader;
    procedure HeaderWndProc(var Message: TMessage);
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMNotify(var Message: TWMNotify); message WM_NOTIFY;
    procedure WMParentNotify(var Message: TWMParentNotify); message WM_PARENTNOTIFY;
    function GetContextMenu(out ContextMenu: IContextMenu): Boolean;
    procedure SetShellTree(Value: TCustomShellTree);
    procedure Editing;
    function GetFilter: string;
    procedure SetFilter(const Value: string);
    procedure DoColumnMenuClick(Sender: TObject);
    procedure AddColumn(const Name: string; Kind: TColumnKind);
    procedure InitColumns;
    procedure StoreColumnsWidth;
    procedure DoDestroyWnd;
    function InCutList(PIDL: PItemIdList): Boolean;
    procedure NotifyClipboard;
    procedure FileNotifyStart;
    procedure FileNotifyStop;
    procedure SaveState;
    procedure RestoreState(VisibleFocused: Boolean);
    procedure InternalLoadColumns(IniFile: TObject; const Section: string);
    procedure InternalSaveColumns(IniFile: TObject; const Section: string);
    procedure ClearDropTargetItem;
    procedure OleDragDrop(Index: Integer; IsRightButton: Boolean);
    function GetDropTarget(MousePos: TPoint): IDropTarget;
    function DragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; stdcall;
    function IDropTarget_DragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;
    function IDropTarget.DragOver = IDropTarget_DragOver;
    function DragLeave: HResult; stdcall;
    function Drop(const dataObj: IDataObject; grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;
  protected
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure Loaded; override;
    function OwnerDataFetch(Item: TListItem; Request: TItemRequest): Boolean; override;
    function OwnerDataFind(Find: TItemFind; const FindString: string;
      const FindPosition: TPoint; FindData: Pointer; StartIndex: Integer;
      Direction: TSearchDirection; Wrap: Boolean): Integer; override;
    procedure ColClick(Column: TListColumn); override;
    procedure DoInfoTip(Item: TListItem; var InfoTip: string); override;
    procedure DoContextPopup(MousePos: TPoint; var Handled: Boolean); override;
    procedure WndProc(var Message: TMessage); override;
    function CanEdit(Item: TListItem): Boolean; override;
    procedure Edit(const Item: TLVItem); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DblClick; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetRootPIDL(PIDL: PItemIDList);
    procedure DoVerb(Verb: PChar);
    procedure Rename(Index: Integer; const NewName: string);
    procedure Execute(Index: Integer);
    procedure ExecuteSelected;
    function GetItemPath(Index: Integer): string;
    function CanPaste: Boolean;
    procedure RemoveColumn(Kind: TColumnKind);
    procedure LoadColumns(IniFile: TRegIniFile; const Section: string); overload;
    procedure LoadColumns(IniFile: TIniFile; const Section: string); overload;
    procedure SaveColumns(IniFile: TRegIniFile; const Section: string); overload;
    procedure SaveColumns(IniFile: TIniFile; const Section: string); overload;
    procedure UpdateRoot;
{$IFNDEF AS_VCL6}
    procedure SelectAll;
{$ENDIF}
    property AutoContext: Boolean read FAutoContext write FAutoContext;
    property Filter: string read GetFilter write SetFilter;
    property Items;
  end;

  TShellList = class(TCustomShellList)
  published
    property AutoContext default True;
    property Filter;
    property Align;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property Color;
    property Constraints;
    property Ctl3D;
    property Enabled;
    property Font;
    property FlatScrollBars;
    property GridLines;
    property HotTrack;
    property HotTrackStyles;
    property HoverTime;
    property OwnerDraw;
    property ReadOnly default False;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
    property OnAdvancedCustomDrawSubItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnContextPopup;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnCustomDrawSubItem;
    property OnDblClick;
    property OnDrawItem;
    property OnEditing;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnSelectItem;
  end;

type
  PShellItem = ^TShellItem;
  TShellItem = packed record
    PIDL: PItemIDList;
    ImageIndex: Integer;
    DisplayName: string;
    Size: Int64;
    TypeName: string;
    Modified: TDateTime;
    Cut: Boolean;
  end;

function EnableAutoComplete(Handle: THandle; FileSystem, URL: Boolean): Boolean;

implementation

uses
  AS_ShellUtils, SysUtils, ComObj, Forms, TypInfo, Math;

{ Undocumented SHChangeNotify API }

type
  PNotifyRegister = ^TNotifyRegister;
  TNotifyRegister = packed record
    pidlPath: PItemIDList;
    bWatchSubtree: BOOL;
  end;

type
  PNotifyItem = ^TNotifyItem;
  TNotifyItem = packed record
    dwItem1: Pointer;
    dwItem2: Pointer;
  end;

const
  WM_SHCHANGENOTIFY = WM_USER + 101;

const
  shell32 = 'shell32.dll';

function SHChangeNotifyRegister(hWnd: HWND; dwFlags: DWORD; wEventMask: Cardinal;
  uMsg: UINT; cItems: UINT; lpItems: PNotifyRegister): THandle; stdcall; external shell32 index 2;

function SHChangeNotifyDeregister(hNotify: THandle): BOOL; stdcall; external shell32 index 4;

{ Common Routines }

function GetShellFolder(PIDL: PItemIDList; Check: Boolean = True): IShellFolder;
var
  HR: HResult;
begin
  if IsDesktopPIDL(PIDL) then
    Result := DesktopFolder
  else
  begin
    Result := nil;
    HR := DesktopFolder.BindToObject(PIDL, nil, IID_IShellFolder,
      Pointer(Result));
    if Check then
      OLECheck(HR);
  end;
end;

function GetParentFolder(PIDL: PItemIDList): IShellFolder;
var
  PrevPIDL: PItemIDList;
begin
  PrevPIDL := CopyPIDL(PIDL);
  try
    StripLastPIDL(PrevPIDL);
    Result := GetShellFolder(PrevPIDL);
  finally
    FreePIDL(PrevPIDL);
  end;
end;

{$IFNDEF AS_VCL6}

function DirectoryExists(const Directory: string): Boolean;
var
  Code: Integer;
begin
  Code := GetFileAttributes(PChar(Directory));
  Result := (Code <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code <> 0);
end;
{$ENDIF}

function IsValidFolder(PIDL: PItemIDList): Boolean;

  function IsValidDirectory(const PathName: string): Boolean;
  begin
    Result := (PathName[Length(PathName)] = '\') or DirectoryExists(PathName);
  end;

begin
  if GetShellFolder(PIDL, False) <> nil then
    Result := not HasPIDLAttributes(PIDL, SFGAO_FILESYSTEM) or
      IsValidDirectory(GetPIDLPath(PIDL))
  else
    Result := False;
end;

function IsDrivePIDL(PIDL: PItemIDList): Boolean;

  function CheckPath(const PathName: string): Boolean;
  begin
    Result := PathName[Length(PathName)] = '\';
  end;

begin
  Result := HasPIDLAttributes(PIDL, SFGAO_FILESYSTEM) and
    CheckPath(GetPIDLPath(PIDL));
end;

function GetDescriptionID(ParentFolder: IShellFolder;
  RelativePIDL: PItemIDList): Cardinal;
var
  SHDescriptionID: TSHDescriptionID;
begin
  FillChar(SHDescriptionID, SizeOf(SHDescriptionID), 0);
  if SHGetDataFromIDList(
    ParentFolder,
    RelativePIDL,
    SHGDFIL_DESCRIPTIONID,
    @SHDescriptionID,
    SizeOf(SHDescriptionID)
    ) = S_OK then
    Result := SHDescriptionID.dwDescriptionId
  else
    Result := 0;
end;

procedure RemoveLinkVerb(Menu: HMenu);
var
  I: Integer;
  MenuItemInfo: TMenuItemInfo;
  Buffer: array[0..79] of Char;
  PrevSeparator: Boolean;
begin
  DeleteMenu(Menu, 17, MF_BYCOMMAND);
  PrevSeparator := False;
  for I := Pred(GetMenuItemCount(Menu)) downto 0 do
  begin
    MenuItemInfo.cbSize := 44;
    MenuItemInfo.fMask := MIIM_TYPE;
    MenuItemInfo.dwTypeData := Buffer;
    MenuItemInfo.cch := SizeOf(Buffer);
    if GetMenuItemInfo(Menu, I, True, MenuItemInfo) and
      ValueInFlags(MFT_SEPARATOR, MenuItemInfo.fType) then
    begin
      if PrevSeparator then
      begin
        DeleteMenu(Menu, I, MF_BYPOSITION);
        Break;
      end
      else
        PrevSeparator := True;
    end
    else
      PrevSeparator := False;
  end;
end;

const
  CMD_OFFSET = 1;

function ShowContextMenu(Handle: THandle; ContextMenu: IContextMenu;
  var ContextMenu2: IContextMenu2; MousePos: TPoint; CanRename: Boolean): Cardinal;
var
  Flags, ErrorMode: Cardinal;
  Menu: HMenu;
begin
  Result := 0;
  Flags := CMF_EXPLORE;
  if CanRename then
    Flags := Flags or CMF_CANRENAME;
  Windows.ClientToScreen(Handle, MousePos);
  Menu := CreatePopupMenu;
  if Menu = 0 then
    Exit;
  try
    ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
    try
      ContextMenu.QueryContextMenu(Menu, 0, CMD_OFFSET, $7FFF, Flags);
    finally
      SetErrorMode(ErrorMode);
    end;
    RemoveLinkVerb(Menu);
    ContextMenu.QueryInterface(IID_IContextMenu2, Pointer(ContextMenu2));
    try
      Result := Cardinal(TrackPopupMenu(Menu, TPM_LEFTALIGN or TPM_RETURNCMD or
        TPM_RIGHTBUTTON, MousePos.X, MousePos.Y, 0, Handle, nil));
      if Result > CMD_OFFSET then
        Dec(Result, CMD_OFFSET)
      else
        Result := 0;
    finally
      ContextMenu2 := nil;
    end;
  finally
    DestroyMenu(Menu);
  end;
end;

function ContextMenuCommand(ContextMenu: IContextMenu; idCmd: Cardinal): string;
var
  ZVerb: array[0..255] of Char;
begin
  ZVerb[0] := #0;
  if ContextMenu.GetCommandString(idCmd, GCS_VERB, nil,
    ZVerb, SizeOf(ZVerb)) = S_OK then
    Result := ZVerb
  else
    Result := '';
end;

procedure ContextMenuVerb(Handle: THandle; ContextMenu: IContextMenu; Verb: PChar);
var
  CMInvokeCommandInfo: TCMInvokeCommandInfo;
begin
  FillChar(CMInvokeCommandInfo, SizeOf(CMInvokeCommandInfo), 0);
  with CMInvokeCommandInfo do
  begin
    cbSize := SizeOf(CMInvokeCommandInfo);
    hwnd := Handle;
    lpVerb := Verb;
    nShow := SW_SHOWNORMAL;
  end;
  ContextMenu.InvokeCommand(CMInvokeCommandInfo);
end;

procedure ContextMenuHint(ContextMenu: IContextMenu; Msg: TWMMenuSelect);
var
  idCmd: Cardinal;
  ZVerb: array[0..255] of Char;
  S: string;
begin
  S := #9;
  if Msg.MenuFlag and (MF_POPUP or MF_SEPARATOR) = 0 then
  begin
    ZVerb[0] := #0;
    idCmd := Msg.IDItem - CMD_OFFSET;
    if (ContextMenu.GetCommandString(idCmd, GCS_HELPTEXT, nil,
      ZVerb, SizeOf(ZVerb)) = S_OK) and (ZVerb[1] <> #0) then
      S := ZVerb;
  end;
  Application.Hint := S;
end;

{ TDropSource }

type
  TDropSource = class(TInterfacedObject, IDropSource)
  private
    FDragButton: Longint;
  protected
    function QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult; stdcall;
    function GiveFeedback(dwEffect: Longint): HResult; stdcall;
  public
    constructor Create(IsRightButton: Boolean);
  end;

constructor TDropSource.Create(IsRightButton: Boolean);
begin
  if IsRightButton then
    FDragButton := MK_RBUTTON
  else
    FDragButton := MK_LBUTTON;
end;

function TDropSource.GiveFeedback(dwEffect: Longint): HResult;
begin
  Result := DRAGDROP_S_USEDEFAULTCURSORS;
end;

function TDropSource.QueryContinueDrag(fEscapePressed: BOOL; grfKeyState: Longint): HResult;
const
  BUTTONS = (MK_LBUTTON or MK_MBUTTON or MK_RBUTTON);
begin
  if fEscapePressed or (((FDragButton xor BUTTONS) and grfKeyState) <> 0) then
    Result := DRAGDROP_S_CANCEL
  else if ((FDragButton and grfKeyState) = 0) then
    Result := DRAGDROP_S_DROP
  else
    Result := S_OK;
end;

{ TFileNotifyThread }

type
  TFileNotifyThread = class(TThread)
  private
    FMutex: THandle;
    FWaitHandle: THandle;
    FPathName: string;
    FWatchSubtree: Boolean;
    FChangeEvent: TThreadMethod;
  protected
    procedure Execute; override;
  public
    constructor Create(const PathName: string; WatchSubtree: Boolean;
      ChangeEvent: TThreadMethod);
    destructor Destroy; override;
    procedure Stop;
  end;

constructor TFileNotifyThread.Create(const PathName: string;
  WatchSubtree: Boolean; ChangeEvent: TThreadMethod);
begin
  FreeOnTerminate := True;
  FPathName := PathName;
  FWatchSubtree := WatchSubtree;
  FChangeEvent := ChangeEvent;
  FMutex := CreateMutex(nil, True, nil);
  inherited Create(True);
end;

destructor TFileNotifyThread.Destroy;
begin
  if FWaitHandle <> ERROR_INVALID_HANDLE then
    FindCloseChangeNotification(FWaitHandle);
  CloseHandle(FMutex);
  inherited Destroy;
end;

procedure TFileNotifyThread.Execute;
var
  Handles: array[0..1] of THandle;
  ObjectID: Cardinal;
begin
  FWaitHandle :=
    FindFirstChangeNotification(PChar(FPathName), FWatchSubtree,
    FILE_NOTIFY_CHANGE_FILE_NAME or FILE_NOTIFY_CHANGE_ATTRIBUTES or
    FILE_NOTIFY_CHANGE_SIZE or FILE_NOTIFY_CHANGE_LAST_WRITE);
  if FWaitHandle = ERROR_INVALID_HANDLE then
    Exit;
  while not Terminated do
  begin
    Handles[0] := FWaitHandle;
    Handles[1] := FMutex;
    ObjectID := WaitForMultipleObjects(2, @Handles, False, INFINITE);
    case ObjectID of
      WAIT_OBJECT_0:
        begin
          if not DirectoryExists(FPathName) then
            Exit;
          Synchronize(FChangeEvent);
          FindNextChangeNotification(FWaitHandle);
        end;
      WAIT_OBJECT_0 + 1, WAIT_FAILED:
        Exit;
    end;
  end;
end;

procedure TFileNotifyThread.Stop;
begin
  ReleaseMutex(FMutex);
end;

{ TCustomShellTree }

function GetDesktopRoot: string;
begin
  Result := GetEnumName(TypeInfo(TRootSpecial), Ord(rsDesktop));
end;

constructor TCustomShellTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRoot := GetDesktopRoot;
  FHistoryList := nil;
  OnCompare := DoCompare;
  ChangeDelay := 500;
  RightClickSelect := True;
  FAutoContext := True;
  FContextMenu2 := nil;
  FCutPIDL := nil;
  FCutList := TList.Create;
  FShellList := nil;
end;

type
  PHistoryItem = ^THistoryItem;
  THistoryItem = packed record
    Root: TRoot;
    PIDL: PItemIDList;
  end;

procedure ClearHistoryItem(HistoryItem: PHistoryItem);
begin
  FreePIDL(HistoryItem.PIDL);
  Dispose(HistoryItem);
end;

destructor TCustomShellTree.Destroy;
var
  I: Integer;
begin
  for I := 0 to Pred(FCutList.Count) do
    FreePIDL(FCutList[I]);
  FCutList.Free;
  FreePIDL(FCutPIDL);
  if FHistoryList <> nil then
  begin
    for I := 0 to Pred(FHistoryList.Count) do
      ClearHistoryItem(FHistoryList[I]);
    FHistoryList.Free;
  end;
  inherited Destroy;
end;

procedure TCustomShellTree.CreateWnd;
var
  NotifyRegister: TNotifyRegister;
begin
  inherited CreateWnd;
  TreeView_SetImageList(Handle, GetSystemImageListHandle(False), TVSIL_NORMAL);
  RootChanged;
  NotifyRegister.pidlPath := nil;
  NotifyRegister.bWatchSubtree := True;
  FNotifyHandle :=
    SHChangeNotifyRegister(Handle, SHCNF_TYPE or SHCNF_IDLIST,
    SHCNE_ALLEVENTS or SHCNE_INTERRUPT, WM_SHCHANGENOTIFY, 1, @NotifyRegister);
  FNextViewerHandle := SetClipboardViewer(Handle);
  FClipboardViewer := True;
  FRegisterDragDrop := RegisterDragDrop(Handle, Self) = S_OK;
end;

procedure TCustomShellTree.DestroyWnd;
begin
  DoDestroyWnd;
  if Items.Count > 0 then
    Items.Clear;
  inherited DestroyWnd;
end;

procedure TCustomShellTree.WMDestroy(var Message: TWMDestroy);
begin
  DoDestroyWnd;
  inherited;
end;

procedure TCustomShellTree.DoDestroyWnd;
begin
  if FClipboardViewer then
  begin
    if HandleAllocated then
      ChangeClipboardChain(Handle, FNextViewerHandle);
    FNextViewerHandle := 0;
    FClipboardViewer := False;
  end;
  if FRegisterDragDrop then
  begin
    if HandleAllocated then
      RevokeDragDrop(Handle);
    FRegisterDragDrop := False;
  end;
  if FNotifyHandle <> 0 then
  begin
    SHChangeNotifyDeregister(FNotifyHandle);
    FNotifyHandle := 0;
  end;
end;

procedure TCustomShellTree.Delete(Node: TTreeNode);
begin
  FreePIDL(Node.Data);
  inherited Delete(Node);
end;

procedure TCustomShellTree.Loaded;
begin
  inherited Loaded;
  RootChanged;
end;

procedure TCustomShellTree.RootChanged;
begin
  if FUpdating or (csLoading in ComponentState) then
    Exit;
  FUpdating := True;
  try
    Items.BeginUpdate;
    try
      CreateRoot;
      FastSelect(Items.GetFirstNode);
    finally
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.CreateRoot;

  function CreatePIDLFromRoot(const OldRoot: TRoot; var NewRoot: TRoot): PItemIDList;
  var
    OrdValue: Integer;
  begin
    try
      OrdValue := GetEnumValue(TypeInfo(TRootSpecial), NewRoot);
      if OrdValue >= 0 then
        Result := CreateSpecialPIDL(TRootSpecial(OrdValue))
      else
      begin
        Result := PathToPIDL(NewRoot);
      end;
    except
      NewRoot := OldRoot;
      raise;
    end;
  end;

var
  NewPIDL: PItemIDList;
  RootNode: TTreeNode;
begin
  NewPIDL := CreatePIDLFromRoot(FOldRoot, FRoot);
  if Items.Count > 0 then
    Items.Clear;
  ShowRoot := not IsDesktopPIDL(NewPIDL);
  RootNode := Items.AddObject(nil, '', NewPIDL);
  DoAdded(RootNode);
  RootNode.Expand(False);
end;

function TCustomShellTree.CanExpand(Node: TTreeNode): Boolean;
begin
  Result := True;
  if (csDesigning in ComponentState) and (Node.Level > 0) then
    Exit;
  if Assigned(OnExpanding) then
    OnExpanding(Self, Node, Result);
  if Result and (Node.HasChildren) and (Node.Count = 0) and
    not PopulateNode(Node) then
  begin
    Result := False;
    Node.HasChildren := True;
    Exit;
  end;
  Node.HasChildren := Node.Count > 0;
end;

procedure TCustomShellTree.SetRoot(const Value: TRoot);
begin
  if AnsiCompareText(FRoot, Value) <> 0 then
  begin
    FOldRoot := FRoot;
    FRoot := Value;
    RootChanged;
  end;
end;

function TCustomShellTree.InCutList(PIDL: PItemIdList): Boolean;
var
  I: Integer;
begin
  for I := 0 to Pred(FCutList.Count) do
    if SamePIDL(FCutList[I], PIDL) then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

procedure TCustomShellTree.DoAdded(Node: TTreeNode);
var
  PIDL: PItemIDList;
  Attributes: Cardinal;
  ParentNode: TTreeNode;
begin
  PIDL := Node.Data;
  Attributes := GetPIDLAttributes(PIDL, SFGAO_HASSUBFOLDER or SFGAO_SHARE);
  Node.ImageIndex := GetPIDLImage(PIDL, False, False);
  Node.SelectedIndex := GetPIDLImage(PIDL, False, True);
  Node.Text := GetPIDLDisplayName(PIDL);
  Node.HasChildren := ValueInFlags(SFGAO_HASSUBFOLDER, Attributes);
  if ValueInFlags(SFGAO_SHARE, Attributes) then
    Node.OverlayIndex := 0;
  ParentNode := Node.Parent;
  if ParentNode = nil then
  begin
    if InCutList(PIDL) then
      Node.Cut := True;
  end
  else
  begin
    if SamePIDL(FCutPIDL, ParentNode.Data) and InCutList(PIDL) then
      Node.Cut := True;
  end;
end;

function TCustomShellTree.PopulateNode(Node: TTreeNode): Boolean;
var
  ParentPIDL, RelativePIDL, NewPIDL: PItemIDList;
  ShellFolder: IShellFolder;
  EnumIDList: IEnumIDList;
  NumIDs, Attributes, ErrorMode: Cardinal;
  SaveCursor: TCursor;
  OwnerWindow: THandle;
  NewNode: TTreeNode;
begin
  SaveCursor := Screen.Cursor;
  Items.BeginUpdate;
  try
    Screen.Cursor := crHourglass;
    Result := False;
    ParentPIDL := Node.Data;
    ShellFolder := GetShellFolder(ParentPIDL);
    ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
    try
      if FDropMode then
        OwnerWindow := 0
      else
        OwnerWindow := Application.Handle;
      Result := ShellFolder.EnumObjects(OwnerWindow, SHCONTF_FOLDERS,
        EnumIDList) = S_OK;
    finally
      SetErrorMode(ErrorMode);
    end;
    if not Result then
      Exit;
    NewPIDL := nil;
    while EnumIDList.Next(1, RelativePIDL, NumIDs) = S_OK do
    try
      NewPIDL := ConcatPIDLs(ParentPIDL, RelativePIDL);
      Attributes :=
        GetPIDLAttributes(NewPIDL, SFGAO_FILESYSTEM or SFGAO_FILESYSANCESTOR);
      if ValueInFlags(SFGAO_FILESYSTEM, Attributes) or
        ValueInFlags(SFGAO_FILESYSANCESTOR, Attributes) then
      begin
        NewNode := Items.AddChildObject(Node, '', NewPIDL);
        NewPIDL := nil;
        DoAdded(NewNode);
      end;
    finally
      FreePIDL(NewPIDL);
      FreePIDL(RelativePIDL);
    end;
    Node.AlphaSort;
  finally
    Screen.Cursor := SaveCursor;
    Items.EndUpdate;
  end;
end;

procedure TCustomShellTree.DoCompare(Sender: TObject; Node1,
  Node2: TTreeNode; Data: Integer; var Compare: Integer);
begin
  Compare := SmallInt(DesktopFolder.CompareIDs(0, Node1.Data, Node2.Data));
end;

procedure TCustomShellTree.AddHistoryItem(PIDL: PItemIDList);
var
  HistoryItem: PHistoryItem;
begin
  if (FHistoryList = nil) or FLockHistory then
    Exit;

  if FHistoryList.Count > 0 then
  begin
    HistoryItem := FHistoryList.Last;
    if (AnsiCompareText(HistoryItem.Root, Root) = 0) and
      SamePIDL(HistoryItem.PIDL, PIDL) then
      Exit;
  end;

  begin
    HistoryItem := New(PHistoryItem);
    FillChar(HistoryItem^, SizeOf(THistoryItem), 0);
    HistoryItem.Root := Root;
    HistoryItem.PIDL := CopyPIDL(PIDL);
    FHistoryList.Add(HistoryItem);
    DoHistory;
  end;
end;

procedure TCustomShellTree.StartHistory;
var
  Node: TTreeNode;
begin
  if FHistoryList = nil then
  begin
    FHistoryList := TList.Create;
    Node := Selected;
    if Node <> nil then
      AddHistoryItem(Node.Data);
  end;
end;

procedure TCustomShellTree.DoHistory;
begin
  if Assigned(FOnHistory) then
    FOnHistory(Self);
end;

procedure TCustomShellTree.HistoryBack;
var
  HistoryItem: PHistoryItem;
begin
  if not IsHistory then
    Exit;
  if FHistoryList.Count > 1 then
  begin
    HistoryItem := FHistoryList.Last;
    ClearHistoryItem(HistoryItem);
    FHistoryList.Delete(Pred(FHistoryList.Count));
  end;
  while FHistoryList.Count > 0 do
  begin
    HistoryItem := FHistoryList.Last;
    if IsValidFolder(HistoryItem.PIDL) then
      Break;
    ClearHistoryItem(HistoryItem);
    FHistoryList.Delete(Pred(FHistoryList.Count));
  end;
  if FHistoryList.Count > 0 then
  begin
    HistoryItem := FHistoryList.Last;
    FLockHistory := True;
    try
      try
        Root := HistoryItem.Root;
      except
        HistoryItem.Root := GetDesktopRoot;
        Root := HistoryItem.Root;
      end;
      SelectFromPIDL(HistoryItem.PIDL);
    finally
      FLockHistory := False;
    end;
    DoHistory;
  end;
end;

function TCustomShellTree.IsHistory: Boolean;
begin
  Result := (FHistoryList <> nil) and (FHistoryList.Count > 1);
end;

function TCustomShellTree.CanChange(Node: TTreeNode): Boolean;
begin
  Result := inherited CanChange(Node);
end;

procedure TCustomShellTree.UpdateShellList(Node: TTreeNode);
begin
  if FShellList <> nil then
  begin
    if Node = nil then
      FShellList.SetRootPIDL(nil)
    else
      FShellList.SetRootPIDL(Node.Data);
  end;
end;

procedure TCustomShellTree.Change(Node: TTreeNode);
begin
  inherited Change(Node);
  if Node <> nil then
    AddHistoryItem(Node.Data);
  UpdateShellList(Node);
end;

procedure TCustomShellTree.DoNodeFromPIDL(PIDL: PItemIDList; out Node: TTreeNode; LastFound: Boolean);

  function ChildExists(FindPIDL: PItemIDList; InNode: TTreeNode): TTreeNode;
  begin
    Result := InNode.GetFirstChild;
    while Result <> nil do
    begin
      if SamePIDL(FindPIDL, Result.Data) then
        Break;
      Result := InNode.GetNextChild(Result);
    end;
  end;

  function CreatePIDLList(PIDL: PItemIDList): TList;
  var
    TempPIDL: PItemIDList;
  begin
    Result := TList.Create;
    TempPIDL := CopyPIDL(PIDL);
    try
      while TempPIDL^.mkid.cb <> 0 do
      begin
        Result.Insert(0, CopyPIDL(TempPIDL));
        StripLastPIDL(TempPIDL);
      end;
    finally
      FreePIDL(TempPIDL);
    end;
  end;

  procedure DestroyPIDLList(List: TList);
  var
    I: Integer;
  begin
    for I := 0 to Pred(List.Count) do
      FreePIDL(List[I]);
    List.Free;
  end;

var
  I, Start: Integer;
  PIDLs: TList;
  ChildNode: TTreeNode;
begin
  Node := Items.GetFirstNode;
  if (Node <> nil) and not SamePIDL(PIDL, Node.Data) then
  begin
    PIDLs := CreatePIDLList(PIDL);
    try
      Start := GetPIDLCount(Node.Data);
      for I := Start to Pred(PIDLs.Count) do
      begin
        if LastFound then
          Node.Expand(False);
        ChildNode := ChildExists(PIDLs[I], Node);
        if ChildNode = nil then
          Break;
        Node := ChildNode;
      end;
    finally
      DestroyPIDLList(PIDLs);
    end;
  end;
end;

procedure TCustomShellTree.DoNodeFromPIDL(PIDL: PItemIDList; out Node, ParentNode: TTreeNode);
var
  ParentPIDL: PItemIDList;
begin
  DoNodeFromPIDL(PIDL, Node, False);
  if SamePIDL(Node.Data, PIDL) then
    ParentNode := Node.Parent
  else
  begin
    ParentPIDL := CopyPIDL(PIDL);
    try
      StripLastPIDL(ParentPIDL);
      if SamePIDL(Node.Data, ParentPIDL) then
        ParentNode := Node
      else
        ParentNode := nil;
    finally
      FreePIDL(ParentPIDL);
    end;
    Node := nil;
  end;
end;

procedure TCustomShellTree.DoSelectFromPIDL(PIDL: PItemIDList);
var
  Node: TTreeNode;
begin
  DoNodeFromPIDL(PIDL, Node, True);
  FastSelect(Node);
  Node.MakeVisible;
end;

procedure TCustomShellTree.SelectFromPIDL(PIDL: PItemIDList);
var
  Node: TTreeNode;
begin
  if FUpdating or (csLoading in ComponentState) or not Assigned(PIDL) then
    Exit;
  HandleNeeded;
  FUpdating := True;
  try
    Node := Selected;
    if (Node <> nil) and SamePIDL(Node.Data, PIDL) then
      Exit;
    Items.BeginUpdate;
    try
      Node := Items.GetFirstNode;
      if (Node = nil) or not IsChildPIDL(Node.Data, PIDL) then
      begin
        FRoot := GetDesktopRoot;
        CreateRoot;
      end;
      DoSelectFromPIDL(PIDL);
    finally
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
end;

function TCustomShellTree.NameForAddressBar: string;
var
  Node: TTreeNode;
begin
  Node := Selected;
  if Node = nil then
    Result := ''
  else
    Result := GetPIDLNameForAddressBar(Node.Data);
end;

function TCustomShellTree.GetPath: string;
var
  Node: TTreeNode;
begin
  Node := Selected;
  if Node = nil then
    Result := ''
  else
    Result := GetPIDLPath(Node.Data);
end;

procedure TCustomShellTree.SetPath(const Value: string);
var
  NewPIDL: PItemIDList;
begin
  NewPIDL := PathToPIDL(Value);
  try
    SelectFromPIDL(NewPIDL);
  finally
    FreePIDL(NewPIDL);
  end;
end;

function TCustomShellTree.GetBookmark: string;
var
  Node: TTreeNode;
begin
  Node := Selected;
  if Node = nil then
    Result := ''
  else
    Result := PIDLToString(Node.Data);
end;

procedure TCustomShellTree.SetBookmark(const Value: string);
var
  NewPIDL: PItemIDList;
begin
  NewPIDL := StringToPIDL(Value);
  try
    SelectFromPIDL(NewPIDL);
  finally
    FreePIDL(NewPIDL);
  end;
end;

procedure TCustomShellTree.DoBackRoot(Node: TTreeNode);
var
  PIDL: PItemIDList;
begin
  if Node <> nil then
    PIDL := CopyPIDL(Node.Data)
  else
    PIDL := nil;
  try
    FRoot := GetDesktopRoot;
    CreateRoot;
    if PIDL <> nil then
      DoSelectFromPIDL(PIDL)
    else
      FastSelect(Items.GetFirstNode);
  finally
    FreePIDL(PIDL);
  end;
end;

procedure TCustomShellTree.FastSelect(Node: TTreeNode);
var
  SaveChangeDelay: Integer;
begin
  SaveChangeDelay := ChangeDelay;
  ChangeDelay := 0;
  Selected := Node;
  ChangeDelay := SaveChangeDelay;
end;

procedure TCustomShellTree.DeleteFromParent(Node, ParentNode: TTreeNode);
begin
  if Node.Selected then
    FastSelect(ParentNode);
  if ParentNode.Count = 1 then
    ParentNode.DeleteChildren
  else
    Node.Delete;
end;

function IsPIDLInList(ParentFolder: IShellFolder;
  PIDL: PItemIdList; List: TList): Boolean;
var
  I: Integer;
begin
  for I := 0 to Pred(List.Count) do
    if ParentFolder.CompareIDs(0, List[I], PIDL) = 0 then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

const
  IOverlayIndex: array[Boolean] of Integer = (-1, 0);

procedure TCustomShellTree.DoRefresh(Node: TTreeNode; OnlyExpanded: Boolean);
var
  TestNode, NextNode, NewNode: TTreeNode;
  PIDLs: TList;
  ParentPIDL, RelativePIDL, NewPIDL: PItemIDList;
  ShellFolder: IShellFolder;
  EnumIDList: IEnumIDList;
  NumIDs, Attributes: Cardinal;
  IsEnum, NewAdded: Boolean;
begin
  if Node.Expanded or not OnlyExpanded and (Node.Count > 0) then
  begin
    PIDLs := TList.Create;
    try
      ParentPIDL := Node.Data;
      TestNode := Node.GetFirstChild;
      while TestNode <> nil do
      begin
        NextNode := Node.GetNextChild(TestNode);
        RelativePIDL := LastPIDL(TestNode.Data);
        NewPIDL := ConcatPIDLs(ParentPIDL, RelativePIDL);
        try
          if IsValidFolder(NewPIDL) then
          begin
            PIDLs.Add(LastPIDL(NewPIDL));
            FreePIDL(TestNode.Data);
            TestNode.Data := NewPIDL;
            NewPIDL := nil;
            DoRefresh(TestNode, OnlyExpanded);
          end
          else
            DeleteFromParent(TestNode, Node);
        finally
          FreePIDL(NewPIDL);
        end;
        TestNode := NextNode;
      end;
      ShellFolder := GetShellFolder(ParentPIDL);
      IsEnum := ShellFolder.EnumObjects(0, SHCONTF_FOLDERS,
        EnumIDList) = S_OK;
      if IsEnum then
      begin
        NewAdded := False;
        NewPIDL := nil;
        while EnumIDList.Next(1, RelativePIDL, NumIDs) = S_OK do
        try
          if not IsPIDLInList(ShellFolder, RelativePIDL, PIDLs) then
          begin
            NewPIDL := ConcatPIDLs(ParentPIDL, RelativePIDL);
            Attributes := GetPIDLAttributes(NewPIDL,
              SFGAO_FILESYSTEM or SFGAO_FILESYSANCESTOR);
            if ValueInFlags(SFGAO_FILESYSTEM, Attributes) or
              ValueInFlags(SFGAO_FILESYSANCESTOR, Attributes) then
            begin
              NewNode := Items.AddChildObject(Node, '', NewPIDL);
              NewPIDL := nil;
              NewAdded := True;
              DoAdded(NewNode);
            end;
          end;
        finally
          FreePIDL(NewPIDL);
          NewPIDL := nil;
          FreePIDL(RelativePIDL);
        end;
        if NewAdded then
          Node.AlphaSort;
      end;
    finally
      PIDLs.Free;
    end;
  end
  else if OnlyExpanded and (Node.Count > 0) then
    Node.DeleteChildren;
  Attributes := GetPIDLAttributes(Node.Data, SFGAO_HASSUBFOLDER or SFGAO_SHARE);
  Node.HasChildren := ValueInFlags(SFGAO_HASSUBFOLDER, Attributes);
  Node.OverlayIndex := IOverlayIndex[ValueInFlags(SFGAO_SHARE, Attributes)];
end;

procedure TCustomShellTree.Refresh(Node: TTreeNode; OnlyExpanded: Boolean);
var
  SaveCursor: TCursor;
  Temp: TTreeNode;
begin
  if FUpdating or (csLoading in ComponentState) or (Node = nil) then
    Exit;
  FUpdating := True;
  try
    SaveCursor := Screen.Cursor;
    Items.BeginUpdate;
    try
      Screen.Cursor := crHourglass;
      while not IsValidFolder(Node.Data) do
      begin
        Temp := Node.Parent;
        if Temp = nil then
        begin
          DoBackRoot(Node);
          Exit;
        end;
        Node := Temp;
      end;
      DoRefresh(Node, OnlyExpanded);
    finally
      Screen.Cursor := SaveCursor;
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
  UpdateShellList(Selected);
end;

function TCustomShellTree.CanEdit(Node: TTreeNode): Boolean;

  function CanRename(PIDL: PItemIDList): Boolean;
  var
    DescriptionID, Attributes: Cardinal;
  begin
    if IsDesktopPIDL(PIDL) then
      Result := False
    else if Win32MajorVersion < 5 then
      Result := not IsDrivePIDL(PIDL)
    else
    begin
      DescriptionID := GetDescriptionID(GetParentFolder(PIDL), LastPIDL(PIDL));
      Result := DescriptionID in [SHDID_FS_DIRECTORY, SHDID_ROOT_REGITEM];
    end;
    if Result then
    begin
      Attributes := GetPIDLAttributes(PIDL, SFGAO_CANRENAME);
      Result := ValueInFlags(SFGAO_CANRENAME, Attributes);
    end;
  end;

begin
  if not CanRename(Node.Data) then
  begin
    Result := False;
  end
  else
    Result := inherited CanEdit(Node);
end;

function RenamePIDL(PIDL: PItemIDList; const NewName: WideString): PItemIDList;
var
  ParentPIDL, RelativePIDL: PItemIDList;
  ParentFolder: IShellFolder;
begin
  if IsDesktopPIDL(PIDL) then
  begin
    Result := nil;
    Exit;
  end;
  RelativePIDL := nil;
  ParentPIDL := CopyPIDL(PIDL);
  try
    StripLastPIDL(ParentPIDL);
    ParentFolder := GetShellFolder(ParentPIDL);
    if ParentFolder.SetNameOf(
      Application.Handle,
      LastPIDL(PIDL),
      POleStr(NewName),
      SHGDN_NORMAL,
      RelativePIDL) = S_OK then
      Result := ConcatPIDLs(ParentPIDL, RelativePIDL)
    else
      Result := nil;
  finally
    FreePIDL(ParentPIDL);
    FreePIDL(RelativePIDL);
  end;
end;

procedure TCustomShellTree.DoRenamed(Node: TTreeNode);
begin
  if Assigned(FOnRenamed) then
    FOnRenamed(Self, Node);
  if Node.Selected then
    UpdateShellList(Node);
end;

procedure TCustomShellTree.Rename(Node: TTreeNode; const NewName: string);
var
  RenamedPIDL: PItemIDList;
  SaveCursor: TCursor;
  ParentNode: TTreeNode;
begin
  if FUpdating or (Node = nil) then
    Exit;
  FUpdating := True;
  try
    RenamedPIDL := RenamePIDL(Node.Data, NewName);
    try
      if RenamedPIDL = nil then
      begin
        Exit;
      end;
      SaveCursor := Screen.Cursor;
      Items.BeginUpdate;
      try
        FreePIDL(Node.Data);
        Node.Data := RenamedPIDL;
        RenamedPIDL := nil;
        Node.Text := NewName;
        FastSelect(Node);
        DoRefresh(Node, False);
        ParentNode := Node.Parent;
        if ParentNode <> nil then
        begin
          ParentNode.AlphaSort;
          Node.MakeVisible;
        end;
      finally
        Screen.Cursor := SaveCursor;
        Items.EndUpdate;
      end;
    finally
      FreePIDL(RenamedPIDL);
    end;
    DoRenamed(Node);
  finally
    FUpdating := False;
  end;
end;

function TCustomShellTree.GetNodeFromItem(const Item: TTVItem): TTreeNode;
begin
  Result := nil;
  if Items <> nil then
    with Item do
      if (state and TVIF_PARAM) <> 0 then
        Result := Pointer(lParam)
      else
        Result := Items.GetNode(hItem);
end;

procedure TCustomShellTree.Edit(const Item: TTVItem);
begin
  with Item do
    if pszText <> nil then
      Rename(GetNodeFromItem(Item), pszText);
end;

function InPersonalPIDL(PIDL: PItemIDList): Boolean;
begin
  Result := AnsiPos(GetSpecialFolderPath(rsPersonal), GetPIDLPath(PIDL)) = 1;
end;

procedure TCustomShellTree.DoNotifyWin9xFix;
var
  PIDL: PItemIDList;
  Node: TTreeNode;
begin
  PIDL := PathToPIDL('::');
  try
    DoNodeFromPIDL(PIDL, Node, False);
    if SamePIDL(Node.Data, PIDL) then
      DoRefresh(Node, False);
  finally
    FreePIDL(PIDL);
  end;
end;

function BindPIDL(PIDL: PItemIDList): PItemIDList;
var
  ShellFolder: IShellFolder;
  EnumIDList: IEnumIDList;
  ParentPIDL, RelativePIDL, NewLastPIDL: PItemIDList;
  IsEnum: Boolean;
  NumIDs: Cardinal;
begin
  Result := nil;
  if IsDesktopPIDL(PIDL) then
    Exit;
  ParentPIDL := CopyPIDL(PIDL);
  try
    StripLastPIDL(ParentPIDL);
    ShellFolder := GetShellFolder(ParentPIDL);
    IsEnum := ShellFolder.EnumObjects(0, SHCONTF_FOLDERS,
      EnumIDList) = S_OK;
    if IsEnum then
    begin
      NewLastPIDL := LastPIDL(PIDL);
      while EnumIDList.Next(1, RelativePIDL, NumIDs) = S_OK do
      try
        if ShellFolder.CompareIDs(0, RelativePIDL, NewLastPIDL) = 0 then
        begin
          Result := ConcatPIDLs(ParentPIDL, RelativePIDL);
          Break;
        end;
      finally
        FreePIDL(RelativePIDL);
      end;
    end;
  finally
    FreePIDL(ParentPIDL);
  end;
end;

procedure TCustomShellTree.DoNotifyAdd(Node: TTreeNode; ParentNode: TTreeNode; PIDL: PItemIDList);
var
  NewPIDL: PItemIDList;
  NewNode: TTreeNode;
begin
  if (Node = nil) and (ParentNode <> nil) then
  begin
    if ParentNode.Count > 0 then
    begin
      NewPIDL := BindPIDL(PIDL);
      if NewPIDL = nil then
        Exit;
      NewNode := Items.AddChildObject(ParentNode, '', NewPIDL);
      DoAdded(NewNode);
      ParentNode.AlphaSort;
    end
    else
      ParentNode.HasChildren := True;
  end;
end;

procedure TCustomShellTree.NotifyAdd(PIDL: PItemIDList);
var
  IsWin9xFix: Boolean;
  Node, ParentNode: TTreeNode;
  SaveCursor: TCursor;
begin
  if FUpdating or not Assigned(PIDL) then
    Exit;
  FUpdating := True;
  try
    IsWin9xFix := (Win32MajorVersion < 5) and InPersonalPIDL(PIDL);
    DoNodeFromPIDL(PIDL, Node, ParentNode);
    if (Node = nil) and (ParentNode = nil) and not IsWin9xFix then
      Exit;
    SaveCursor := Screen.Cursor;
    Items.BeginUpdate;
    try
      Screen.Cursor := crHourglass;
      DoNotifyAdd(Node, ParentNode, PIDL);
      if IsWin9xFix then
        DoNotifyWin9xFix;
    finally
      Screen.Cursor := SaveCursor;
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.DoNotifyDel(Node, ParentNode: TTreeNode; out IsBackRoot: Boolean);
begin
  if Node <> nil then
  begin
    if ParentNode = nil then
    begin
      DoBackRoot(Node);
      IsBackRoot := True;
      Exit;
    end
    else
      DeleteFromParent(Node, ParentNode);
  end
  else if ParentNode <> nil then
  begin
    ParentNode.HasChildren := (ParentNode.Count > 0) or
      HasPIDLAttributes(ParentNode.Data, SFGAO_HASSUBFOLDER);
  end;
  IsBackRoot := False;
end;

procedure TCustomShellTree.NotifyDel(PIDL: PItemIDList);
var
  IsBackRoot, IsWin9xFix: Boolean;
  Node, ParentNode: TTreeNode;
  SaveCursor: TCursor;
begin
  if FUpdating or not Assigned(PIDL) then
    Exit;
  FUpdating := True;
  try
    IsWin9xFix := (Win32MajorVersion < 5) and InPersonalPIDL(PIDL);
    DoNodeFromPIDL(PIDL, Node, ParentNode);
    if (Node = nil) and (ParentNode = nil) and not IsWin9xFix then
      Exit;
    SaveCursor := Screen.Cursor;
    Items.BeginUpdate;
    try
      Screen.Cursor := crHourglass;
      DoNotifyDel(Node, ParentNode, IsBackRoot);
      if IsBackRoot then
        Exit;
      if IsWin9xFix then
        DoNotifyWin9xFix;
    finally
      Screen.Cursor := SaveCursor;
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.NotifyRename(OldPIDL, NewPIDL: PItemIDList);
var
  IsMoved, IsBackRoot, IsRenamed, IsWin9xFix: Boolean;
  OldNode, OldParentNode,
    NewNode, NewParentNode,
    Node, ParentNode: TTreeNode;
  SaveCursor: TCursor;
  RenamedPIDL: PItemIDList;
begin
  if FUpdating or not Assigned(OldPIDL) or not Assigned(NewPIDL) then
    Exit;
  FUpdating := True;
  try
    IsMoved := AnsiCompareText(
      GetPIDLDisplayName(OldPIDL),
      GetPIDLDisplayName(NewPIDL)) = 0;
    if IsMoved and IsDrivePIDL(OldPIDL) then
      IsMoved := False;
    IsWin9xFix := (Win32MajorVersion < 5) and
      (InPersonalPIDL(OldPIDL) or InPersonalPIDL(NewPIDL));
    if IsMoved then
    begin
      DoNodeFromPIDL(OldPIDL, OldNode, OldParentNode);
      DoNodeFromPIDL(NewPIDL, NewNode, NewParentNode);
      if (OldNode = nil) and (OldParentNode = nil) and (NewNode = nil) and
        (NewParentNode = nil) and not IsWin9xFix then
        Exit;
      Node := nil;
    end
    else
    begin
      DoNodeFromPIDL(OldPIDL, Node, False);
      if not SamePIDL(Node.Data, OldPIDL) then
        Node := nil;
      if (Node = nil) and not IsWin9xFix then
        Exit;
    end;
    IsRenamed := False;
    SaveCursor := Screen.Cursor;
    Items.BeginUpdate;
    try
      if IsMoved then
      begin
        DoNotifyDel(OldNode, OldParentNode, IsBackRoot);
        if IsBackRoot then
          Exit;
        DoNotifyAdd(NewNode, NewParentNode, NewPIDL);
      end
      else if Node <> nil then
      begin
        RenamedPIDL := BindPIDL(NewPIDL);
        if RenamedPIDL = nil then
          Exit;
        FreePIDL(Node.Data);
        Node.Data := RenamedPIDL;
        Node.Text := GetPIDLDisplayName(Node.Data);
        DoRefresh(Node, False);
        ParentNode := Node.Parent;
        if ParentNode <> nil then
          ParentNode.AlphaSort;
        IsRenamed := True;
      end;
      if IsWin9xFix then
        DoNotifyWin9xFix;
    finally
      Screen.Cursor := SaveCursor;
      Items.EndUpdate;
    end;
    if IsRenamed then
      DoRenamed(Node);
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.NotifyUpdateItem(PIDL: PItemIDList);
var
  Node: TTreeNode;
begin
  if FUpdating or not Assigned(PIDL) then
    Exit;
  FUpdating := True;
  try
    DoNodeFromPIDL(PIDL, Node, False);
    if SamePIDL(Node.Data, PIDL) then
    begin
      Node.Text := GetPIDLDisplayName(Node.Data);
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.NotifyMedia(PIDL: PItemIDList);
var
  SaveCursor: TCursor;
  Node, ParentNode: TTreeNode;
begin
  if FUpdating or not Assigned(PIDL) then
    Exit;
  FUpdating := True;
  try
    DoNodeFromPIDL(PIDL, Node, False);
    if not SamePIDL(Node.Data, PIDL) then
      Exit;
    SaveCursor := Screen.Cursor;
    Items.BeginUpdate;
    try
      Screen.Cursor := crHourglass;
      ParentNode := Node.Parent;
      if ParentNode = nil then
      begin
        FRoot := GetPIDLPath(PIDL);
        CreateRoot;
        FastSelect(Items.GetFirstNode);
      end
      else
      begin
        DeleteFromParent(Node, ParentNode);
        DoNotifyAdd(nil, ParentNode, PIDL);
      end;
    finally
      Screen.Cursor := SaveCursor;
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.NotifyShare(PIDL: PItemIDList);
var
  Node: TTreeNode;
  SaveCursor: TCursor;
begin
  if FUpdating or not Assigned(PIDL) then
    Exit;
  FUpdating := True;
  try
    DoNodeFromPIDL(PIDL, Node, False);
    if SamePIDL(Node.Data, PIDL) then
      Node.OverlayIndex :=
        IOverlayIndex[HasPIDLAttributes(Node.Data, SFGAO_SHARE)];
    if (Win32MajorVersion < 5) and InPersonalPIDL(PIDL) then
    begin
      SaveCursor := Screen.Cursor;
      Items.BeginUpdate;
      try
        Screen.Cursor := crHourglass;
        DoNotifyWin9xFix;
      finally
        Screen.Cursor := SaveCursor;
        Items.EndUpdate;
      end;
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.HandleNotify(WParam, LParam: Integer);
var
  NotifyItem: PNotifyItem;
begin
  NotifyItem := Pointer(WParam);
  case LParam of
    SHCNE_MKDIR:
      NotifyAdd(NotifyItem^.dwItem1);
    SHCNE_RMDIR:
      NotifyDel(NotifyItem^.dwItem1);
    SHCNE_MEDIAINSERTED:
      NotifyMedia(NotifyItem^.dwItem1);
    SHCNE_MEDIAREMOVED:
      NotifyMedia(NotifyItem^.dwItem1);
    SHCNE_DRIVEREMOVED:
      NotifyDel(NotifyItem^.dwItem1);
    SHCNE_DRIVEADD:
      NotifyAdd(NotifyItem^.dwItem1);
    SHCNE_NETSHARE:
      NotifyShare(NotifyItem^.dwItem1);
    SHCNE_NETUNSHARE:
      NotifyShare(NotifyItem^.dwItem1);
    SHCNE_UPDATEITEM:
      NotifyUpdateItem(NotifyItem^.dwItem1);
    SHCNE_RENAMEFOLDER:
      NotifyRename(NotifyItem^.dwItem1, NotifyItem^.dwItem2);
  else

  end;
end;

var
  CF_IDLIST: TClipFormat;

procedure TCustomShellTree.NotifyClipboard;

  procedure RefreshCutPIDLs;
  var
    SaveCursor: TCursor;
    Node, TestNode, NextNode: TTreeNode;
  begin
    TestNode := Items.GetFirstNode;
    if TestNode = nil then
      Exit;
    TestNode.Cut := InCutList(TestNode.Data);
    DoNodeFromPIDL(FCutPIDL, Node, False);
    if SamePIDL(Node.Data, FCutPIDL) and (Node.Count > 0) then
    begin
      SaveCursor := Screen.Cursor;
      try
        Screen.Cursor := crHourglass;
        TestNode := Node.GetFirstChild;
        while TestNode <> nil do
        begin
          NextNode := Node.GetNextChild(TestNode);
          TestNode.Cut := InCutList(TestNode.Data);
          TestNode := NextNode;
        end;
      finally
        Screen.Cursor := SaveCursor;
      end;
    end;
  end;

var
  Data: THandle;
  DataPtr: Pointer;
  ChildCount, I: Integer;
  ItemOffset: UINT;
  ParentPIDL, RelativePIDL: PItemIDList;
begin
  if FUpdating then
    Exit;
  FUpdating := True;
  try
    if FCutPIDL <> nil then
    begin
      for I := 0 to Pred(FCutList.Count) do
        FreePIDL(FCutList[I]);
      FCutList.Clear;
      RefreshCutPIDLs;
      FreePIDL(FCutPIDL);
      FCutPIDL := nil;
    end;
    if not IsClipboardFormatAvailable(CF_IDLIST) then
      Exit;
    if OpenClipboard(Handle) then
    try
      Data := GetClipboardData(CF_IDLIST);
      if Data = 0 then
        Exit;
      DataPtr := GlobalLock(Data);
      if DataPtr = nil then
        Exit;
      try
        ChildCount := PIDA(DataPtr)^.cidl;
        ItemOffset := UINT(DataPtr) + SizeOf(UINT);
        ParentPIDL := Pointer(UINT(DataPtr) + PUINT(ItemOffset)^);
        for I := 0 to Pred(ChildCount) do
        begin
          Inc(ItemOffset, SizeOf(UINT));
          RelativePIDL :=
            Pointer(UINT(DataPtr) + PUINT(ItemOffset)^);
          FCutList.Add(ConcatPIDLs(ParentPIDL, RelativePIDL));
        end;
        FCutPIDL := CopyPIDL(ParentPIDL);
      finally
        GlobalUnlock(Data);
      end;
    finally
      CloseClipboard;
    end;
    if FCutPIDL <> nil then
      RefreshCutPIDLs;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_MENUSELECT:
      if ShowHint and Assigned(FContextMenu2) then
      begin
        ContextMenuHint(FContextMenu2, TWMMenuSelect(Message));
        Exit;
      end;
    WM_INITMENUPOPUP,
      WM_DRAWITEM,
      WM_MENUCHAR,
      WM_MEASUREITEM:
      if Assigned(FContextMenu2) then
      begin
        FContextMenu2.HandleMenuMsg(Message.Msg, Message.WParam, Message.LParam);
        Message.Result := 0;
        Exit;
      end;
    WM_CHANGECBCHAIN:
      begin
        if THandle(Message.WParam) = FNextViewerHandle then
          FNextViewerHandle := THandle(Message.LParam)
        else if FNextViewerHandle <> 0 then
          with Message do
            SendMessage(FNextViewerHandle, Msg, WParam, LParam);
        Exit;
      end;
    WM_DRAWCLIPBOARD:
      begin
        NotifyClipboard;
        with Message do
          SendMessage(FNextViewerHandle, Msg, WParam, LParam);
        Exit;
      end;
    WM_SHCHANGENOTIFY:
      begin
        HandleNotify(Message.WParam, Message.LParam);
        Exit;
      end;
  end;
  inherited WndProc(Message);
end;

procedure TCustomShellTree.Editing(Node: TTreeNode);
begin
  if Node <> nil then
    Node.EditText;
end;

function TCustomShellTree.GetContextMenu(out ContextMenu: IContextMenu;
  PIDL: PItemIDList): Boolean;
var
  ParentFolder: IShellFolder;
  RelativePIDL: PItemIDList;
begin
  ContextMenu := nil;
  if IsDesktopPIDL(PIDL) then
  begin
    Result := False;
    Exit;
  end;
  ParentFolder := GetParentFolder(PIDL);
  RelativePIDL := LastPIDL(PIDL);
  Result := ParentFolder.GetUIObjectOf(Application.Handle, 1, RelativePIDL,
    IID_IContextMenu, nil, Pointer(ContextMenu)) = S_OK;
end;

procedure TCustomShellTree.DoVerb(Verb: PChar);
var
  Node: TTreeNode;
  PIDL: PItemIDList;
  IsNewPIDL: Boolean;
  ContextMenu: IContextMenu;
begin
  Node := Selected;
  if Node = nil then
    Exit;
  PIDL := Node.Data;
  if IsDesktopPIDL(PIDL) and SameText(Verb, 'paste') then
  begin
    IsNewPIDL := True;
    PIDL := CreateSpecialPIDL(rsDesktopDirectory);
  end
  else
    IsNewPIDL := False;
  try
    if not GetContextMenu(ContextMenu, PIDL) then
      Exit;
    ContextMenuVerb(ValidParentForm(Self).Handle, ContextMenu, Verb);
  finally
    if IsNewPIDL then
      FreePIDL(PIDL);
  end;
end;

procedure TCustomShellTree.DoContextPopup(MousePos: TPoint;
  var Handled: Boolean);
var
  Node: TTreeNode;
  ContextMenu: IContextMenu;
  idCmd: Cardinal;
begin
  if not FAutoContext then
  begin
    inherited DoContextPopup(MousePos, Handled);
    Exit;
  end;
  Handled := True;
  Node := Selected;
  if (Node = nil) or not GetContextMenu(ContextMenu, Node.Data) then
    Exit;
  idCmd := ShowContextMenu(Handle, ContextMenu, FContextMenu2,
    MousePos, CanEdit(Node));
  if idCmd <> 0 then
  begin
    if SameText(ContextMenuCommand(ContextMenu, idCmd), 'rename') then
      Editing(Node)
    else
      ContextMenuVerb(ValidParentForm(Self).Handle,
        ContextMenu, MakeIntResource(idCmd));
  end;
end;

procedure TCustomShellTree.BackRoot(Node: TTreeNode);
var
  SaveCursor: TCursor;
begin
  if FUpdating then
    Exit;
  FUpdating := True;
  try
    if Node = nil then
    begin
      Node := Selected;
      if (Node = nil) or (Node.Parent <> nil) or
        IsDesktopPIDL(Node.Data) then
        Exit;
    end
    else if IsDesktopPIDL(Items[0].Data) then
      Exit;
    SaveCursor := Screen.Cursor;
    Items.BeginUpdate;
    try
      Screen.Cursor := crHourglass;
      DoBackRoot(Node);
    finally
      Screen.Cursor := SaveCursor;
      Items.EndUpdate;
    end;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellTree.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  case Key of
    VK_INSERT:
      if not IsEditing then
      begin
        NewSubFolder(Selected);
        Key := 0;
      end;
    VK_DELETE:
      if not IsEditing then
      begin
        DoVerb('delete');
        Key := 0;
      end;
    VK_BACK:
      if not IsEditing then
      begin
        if ssCtrl in Shift then
          BackRoot(Selected)
        else
          BackRoot(nil);
      end;
    VK_F2:
      if not IsEditing then
      begin
        Editing(Selected);
        Key := 0;
      end;
    VK_F5:
      if not IsEditing then
      begin
        if ssCtrl in Shift then
        begin
          Refresh(Selected, ssAlt in Shift);
          Key := 0;
        end
        else
        begin
          Refresh(Items.GetFirstNode, ssAlt in Shift);
          Key := 0;
        end;
      end;
  end;
end;

procedure TCustomShellTree.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case Word(Key) of
    Word(^C):
      if not IsEditing then
      begin
        DoVerb('copy');
        Key := #0;
      end;
    Word(^V):
      if not IsEditing then
      begin
        DoVerb('paste');
        Key := #0;
      end;
    Word(^X):
      if not IsEditing then
      begin
        DoVerb('cut');
        Key := #0;
      end;
    127:
      if not IsEditing then
        Key := #0;
  end;
end;

procedure TCustomShellTree.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
end;

const
  AUTOEXPAND_MINDELAY_MS = Cardinal(1000);

function TCustomShellTree.GetDropTarget(MousePos: TPoint): IDropTarget;

  function GetDropTarget(MousePos: TPoint): TTreeNode;
  begin
    with ScreenToClient(MousePos) do
      Result := GetNodeAt(x, y);
    if Self.DropTarget <> Result then
    begin
      Self.DropTarget := Result;
      if (Result <> nil) and not Result.Expanded and Result.HasChildren then
        FStartAutoExpandTick := GetTickCount
      else
        FStartAutoExpandTick := 0;
    end
    else if (FStartAutoExpandTick > 0) and
      (GetTickCount > (FStartAutoExpandTick + AUTOEXPAND_MINDELAY_MS)) then
    begin
      FDropMode := True;
      try
        Result.Expand(False);
      finally
        FDropMode := False;
      end;
      FStartAutoExpandTick := 0;
      Result := GetDropTarget(MousePos);
    end;
  end;

var
  Node: TTreeNode;
  PIDL, RelativePIDL: PItemIDList;
  ParentFolder: IShellFolder;
begin
  Result := nil;
  Node := GetDropTarget(MousePos);
  if Node = nil then
    Exit;
  if IsDesktopPIDL(Node.Data) then
    PIDL := CreateSpecialPIDL(rsDesktopDirectory)
  else
    PIDL := CopyPIDL(Node.Data);
  try
    RelativePIDL := LastPIDL(PIDL);
    ParentFolder := GetParentFolder(PIDL);
    if ParentFolder.GetUIObjectOf(Application.Handle, 1, RelativePIDL,
      IDropTarget, nil, Pointer(Result)) <> S_OK then
      Result := nil;
  finally
    FreePIDL(PIDL);
  end;
end;

const
  AUTOSCROLL_MINDELAY_MS = Cardinal(200);
  AUTOSCROLL_THRESHOLD_X = Integer((20 * 13) div 10);
  AUTOSCROLL_THRESHOLD_Y = Integer(20);

procedure DoAutoScroll(WinControl: TWinControl; MousePos: TPoint;
  var LastTickCount: Cardinal);

  procedure Scroll(const Code, Direction: Integer);
  const
    Conversion: array[SB_HORZ..SB_VERT] of Cardinal = (WM_HSCROLL, WM_VSCROLL);
  var
    ScrollInfo: TScrollInfo;
  begin
    FillChar(ScrollInfo, SizeOf(ScrollInfo), 0);
    with ScrollInfo do
    begin
      cbSize := SizeOf(ScrollInfo);
      fMask := SIF_ALL;
    end;
    GetScrollInfo(WinControl.Handle, Code, ScrollInfo);
    if (ScrollInfo.nMax - ScrollInfo.nMin - Integer(ScrollInfo.nPage)) > 0 then
      WinControl.Perform(Conversion[Code], Direction, 0);
  end;

var
  ClientRect: TRect;
  Point: TPoint;
begin
  if GetTickCount > (LastTickCount + AUTOSCROLL_MINDELAY_MS) then
  begin
    ClientRect := WinControl.ClientRect;
    Point := WinControl.ScreenToClient(MousePos);
    if Point.x < AUTOSCROLL_THRESHOLD_X then
      Scroll(SB_HORZ, SB_LINEUP)
    else if Point.x > (ClientRect.Right - AUTOSCROLL_THRESHOLD_X) then
      Scroll(SB_HORZ, SB_LINEDOWN);
    if Point.y < AUTOSCROLL_THRESHOLD_Y then
      Scroll(SB_VERT, SB_LINEUP)
    else if Point.y > (ClientRect.Bottom - AUTOSCROLL_THRESHOLD_Y) then
      Scroll(SB_VERT, SB_LINEDOWN);
    LastTickCount := GetTickCount;
  end;
end;

function TCustomShellTree.DragEnter(const dataObj: IDataObject;
  grfKeyState: Longint; pt: TPoint; var dwEffect: Longint): HResult;
var
  DropTarget: IDropTarget;
begin
  FLastAutoScrollTick := GetTickCount;
  FDropLeftButton := grfKeyState and MK_LBUTTON <> 0;
  FDataObject := dataObj;
  try
    FStartAutoExpandTick := 0;
    DropTarget := GetDropTarget(pt);
    if DropTarget = nil then
      Abort;
    Result := DropTarget.DragEnter(dataObj, grfKeyState, pt, dwEffect);
    DropTarget.DragLeave;
    if Result <> S_OK then
      Abort;
  except
    dwEffect := DROPEFFECT_NONE;
    Result := S_OK;
  end;
end;

function TCustomShellTree.DragLeave: HResult;
begin
  FDataObject := nil;
  Self.DropTarget := nil;
  Result := S_OK;
end;

function TCustomShellTree.IDropTarget_DragOver(grfKeyState: Longint; pt: TPoint;
  var dwEffect: Longint): HResult;
var
  DropTarget: IDropTarget;
begin
  try
    DoAutoScroll(Self, pt, FLastAutoScrollTick);
    DropTarget := GetDropTarget(pt);
    if DropTarget = nil then
      Abort;
    DropTarget.DragEnter(FDataObject, grfKeyState, pt, dwEffect);
    Result := DropTarget.DragOver(grfKeyState, pt, dwEffect);
    DropTarget.DragLeave;
    if Result <> S_OK then
      Abort;
  except
    dwEffect := DROPEFFECT_NONE;
    Result := S_OK;
  end;
end;

function TCustomShellTree.Drop(const dataObj: IDataObject;
  grfKeyState: Integer; pt: TPoint; var dwEffect: Longint): HResult;
var
  DropTarget: IDropTarget;
  dwEffect2: Longint;
begin
  try
    try
      FStartAutoExpandTick := 0;
      DropTarget := GetDropTarget(pt);
      if DropTarget = nil then
        Abort;
      if FDropLeftButton then
        grfKeyState := grfKeyState or MK_LBUTTON;
      dwEffect2 := dwEffect;
      DropTarget.DragEnter(dataObj, grfKeyState, pt, dwEffect2);
      Result := DropTarget.Drop(dataObj, grfKeyState, pt, dwEffect);
      if Result <> S_OK then
        Abort;
    except
      dwEffect := DROPEFFECT_NONE;
      Result := S_OK;
    end;
  finally
    FDataObject := nil;
    Self.DropTarget := nil;
  end;
end;

procedure TCustomShellTree.OleDragDrop(Node: TTreeNode; IsRightButton: Boolean);
var
  ParentFolder: IShellFolder;
  RelativePIDL: PItemIDList;
  HR: HResult;
  DataObject: IDataObject;
  DropSource: IDropSource;
  Flags: Cardinal;
  Effect: Longint;
begin
  ParentFolder := GetParentFolder(Node.Data);
  RelativePIDL := LastPIDL(Node.Data);
  HR := ParentFolder.GetUIObjectOf(Application.Handle, 1, RelativePIDL,
    IDataObject, nil, Pointer(DataObject));
  if HR = E_NOINTERFACE then
    Exit;
  OleCheck(HR);
  DropSource := TDropSource.Create(IsRightButton);
  Flags := SFGAO_CANCOPY or SFGAO_CANMOVE or SFGAO_CANLINK;
  OleCheck(ParentFolder.GetAttributesOf(1, RelativePIDL, Flags));
  Effect := 0;
  if ValueInFlags(SFGAO_CANCOPY, Flags) then
    Effect := Effect or DROPEFFECT_COPY;
  if ValueInFlags(SFGAO_CANMOVE, Flags) then
    Effect := Effect or DROPEFFECT_MOVE;
  if ValueInFlags(SFGAO_CANLINK, Flags) then
    Effect := Effect or DROPEFFECT_LINK;
  Self.DropTarget := Node;
  DoDragDrop(DataObject, DropSource, Effect, Effect);
  Self.DropTarget := nil;
end;

procedure TCustomShellTree.CNNotify(var Message: TWMNotify);
begin
  with Message do
    case NMHdr^.code of
      TVN_BEGINDRAG, TVN_BEGINRDRAG:
        with PNMTreeView(NMHdr)^ do
          OleDragDrop(GetNodeFromItem(ItemNew), NMHdr^.code = TVN_BEGINRDRAG);
    else
      inherited;
    end;
end;

procedure TCustomShellTree.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style and not TVS_DISABLEDRAGDROP;
end;

procedure TCustomShellTree.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if FChangeTimer.Enabled then
  begin
    FChangeTimer.Enabled := False;
    Change(TTreeNode(FChangeTimer.Tag));
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TCustomShellTree.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TCustomShellTree.SetShellList(Value: TCustomShellList);
begin
  if FShellList <> Value then
  begin
    if FShellList <> nil then
    begin
      FShellList.RemoveFreeNotification(Self);
      FShellList.SetShellTree(nil);
    end;
    FShellList := Value;
    if FShellList <> nil then
    begin
      FShellList.FreeNotification(Self);
      FShellList.SetShellTree(Self);
    end;
  end;
end;

procedure TCustomShellTree.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FShellList then
    begin
      FShellList := nil;
    end;
  end;
end;

var
  NewFolderName: string;

procedure TCustomShellTree.NewSubFolder(Node: TTreeNode);

  procedure RefreshFromPIDL(PIDL: PItemIDList);
  var
    Node: TTreeNode;
    SaveCursor: TCursor;
  begin
    DoNodeFromPIDL(PIDL, Node, False);
    if SamePIDL(Node.Data, PIDL) then
    begin
      SaveCursor := Screen.Cursor;
      Items.BeginUpdate;
      try
        Screen.Cursor := crHourglass;
        DoRefresh(Node, False);
      finally
        Screen.Cursor := SaveCursor;
        Items.EndUpdate;
      end;
    end;
  end;

var
  NewNode: TTreeNode;
  NewPath, S: string;
  N: Integer;
  RootPIDL, TempPIDL, NewPIDL: PItemIDList;
begin
  if FUpdating or (Node = nil) then
    Exit;
  FUpdating := True;
  try
    if not IsFileSystemPIDL(Node.Data) then
    begin
      MessageBeep(0);
      Exit;
    end;
    NewPath :=
      IncludeTrailingBackslash(GetPIDLPath(Node.Data)) + NewFolderName;
    begin
      S := '';
      N := 0;
      repeat
        Inc(N);
        FmtStr(S, '%s (%u)', [NewPath, N]);
      until not DirectoryExists(S);
      NewPath := S;
    end;
    Node.Expand(False);
    if not CreateDirectory(PChar(NewPath), nil) then
      RaiseLastWin32Error;
    RootPIDL := CreateSpecialPIDL(rsDrives);
    try
      if IsChildPIDL(RootPIDL, Node.Data) then
      begin
        if InPersonalPIDL(Node.Data) then
        begin
          FreePIDL(RootPIDL);
          RootPIDL := PathToPIDL('::');
          RefreshFromPIDL(RootPIDL);
        end;
      end
      else
        RefreshFromPIDL(RootPIDL);
    finally
      FreePIDL(RootPIDL);
    end;
    TempPIDL := PathToPIDL(NewPath);
    try
      NewPIDL := ConcatPIDLs(Node.Data, LastPIDL(TempPIDL));
    finally
      FreePIDL(TempPIDL);
    end;
    NewNode := Items.AddChildObjectFirst(Node, '', NewPIDL);
    DoAdded(NewNode);
    Node.Expand(False);
    Editing(NewNode);
  finally
    FUpdating := False;
  end;
end;

function TCustomShellTree.CanPaste: Boolean;
var
  Node: TTreeNode;
begin
  Node := Selected;
  Result := (FCutPIDL <> nil) and (Node <> nil) and IsFileSystemPIDL(Node.Data);
end;

{ TCustomShellList }

procedure TCustomShellList.CNNotify(var Message: TWMNotify);
var
  InfoTip: string;
begin
  with Message do
    case NMHdr^.code of
      LVN_BEGINLABELEDIT:
        begin
          inherited;
          if Result = 0 then
          begin
            FSelectedEdit := Items[PLVDispInfo(NMHdr)^.item.iItem].Selected;
            Selected := nil;
          end;
        end;
      LVN_ENDLABELEDIT:
        begin
          if FSelectedEdit then
            Items[PLVDispInfo(NMHdr)^.item.iItem].Selected := True;
          inherited;
        end;
      LVN_GETINFOTIP:
        begin
          if GetWindowLong(ValidParentForm(Parent).Handle, GWL_EXSTYLE) and
            WS_EX_TOPMOST <> 0 then
          begin
            inherited;
            Exit;
          end;
          with PNMLVGetInfoTip(NMHdr)^ do
          begin
            InfoTip := '';
            DoInfoTip(Items[iItem], InfoTip);
            if (InfoTip <> '') then
              StrLCopy(pszText, PChar(InfoTip), cchTextMax);
          end;
        end;
      LVN_BEGINDRAG, LVN_BEGINRDRAG:
        begin
          OleDragDrop(PNMListView(NMHdr)^.iItem, NMHdr^.code = LVN_BEGINRDRAG);
        end;
    else
      inherited;
    end;
end;

const
  ResIdentifiers: array[TColumnKind] of Integer = (8976, 8978, 8979, 8980);

{$R *.res}

constructor TCustomShellList.Create(AOwner: TComponent);

  procedure InitColumnData(Kind: TColumnKind);
  var
    ColumnName: string;
    MenuItem: TMenuItem;
  begin
    ColumnName := GetShellString(ResIdentifiers[Kind]);
    MenuItem := TMenuItem.Create(nil);
    MenuItem.Caption := ColumnName;
    MenuItem.Enabled := Kind <> ckName;
    MenuItem.Checked := True;
    MenuItem.Tag := Ord(Kind);
    MenuItem.OnClick := DoColumnMenuClick;
    FColumnsMenu.Items.Add(MenuItem);
    FColumnsWidth[Kind] := 50;
  end;

  function CreateBitmap(const Name: string): TBitmap;
  begin
    Result := TBitmap.Create;
    Result.Transparent := True;
    Result.Handle := LoadBitmap(HInstance, PChar(Name));
  end;

var
  KindLoop: TColumnKind;
begin
  inherited Create(AOwner);
  FColumnsMenu := TPopupMenu.Create(Self);
  FColumnsMenu.AutoHotkeys := maManual;
  for KindLoop := Low(TColumnKind) to High(TColumnKind) do
    InitColumnData(KindLoop);
  ColumnClick := True;
  MultiSelect := True;
  ViewStyle := vsReport;
  HideSelection := False;
  RowSelect := True;
  OwnerData := True;
  FShellList := TList.Create;
  FSortColumn := ckName;
  FSortArrows[1] := CreateBitmap('SORT_ARROW_0');
  FSortArrows[0] := CreateBitmap('SORT_ARROW_1');
  FHeaderInstance := MakeObjectInstance(HeaderWndProc);
  FContextMenu2 := nil;
  FAutoContext := True;
  FShellTree := nil;
  FFileNotify := nil;
  FCutPIDL := nil;
  FCutList := TList.Create;
  FSelList := TList.Create;
  FFocusedPIDL := nil;
  FFilter := TStringList.Create;
end;

destructor TCustomShellList.Destroy;
var
  I: Integer;
begin
  FFilter.Free;
  FSelList.Free;
  for I := 0 to Pred(FCutList.Count) do
    FreePIDL(FCutList[I]);
  FCutList.Free;
  FreePIDL(FCutPIDL);
  if FHeaderHandle <> 0 then
    SetWindowLong(FHeaderHandle, GWL_WNDPROC, LongInt(FDefHeaderProc));
  FreeObjectInstance(FHeaderInstance);
  FSortArrows[0].Free;
  FSortArrows[1].Free;
  ClearShellList;
  FShellList.Free;
  inherited Destroy;
end;

procedure TCustomShellList.CreateWnd;
begin
  inherited CreateWnd;
  ListView_SetImageList(Handle, GetSystemImageListHandle(False), LVSIL_SMALL);
  InitColumns;
  FNextViewerHandle := SetClipboardViewer(Handle);
  FClipboardViewer := True;
  FRegisterDragDrop := RegisterDragDrop(Handle, Self) = S_OK;
end;

procedure TCustomShellList.DestroyWnd;
begin
  Items.Count := 0;
  ClearShellList;
  FShellList.Clear;
  StoreColumnsWidth;
  Columns.Clear;
  DoDestroyWnd;
  inherited DestroyWnd;
end;

procedure TCustomShellList.WMDestroy(var Message: TWMDestroy);
begin
  DoDestroyWnd;
  inherited;
end;

procedure TCustomShellList.DoDestroyWnd;
begin
  if FClipboardViewer then
  begin
    if HandleAllocated then
      ChangeClipboardChain(Handle, FNextViewerHandle);
    FNextViewerHandle := 0;
    FClipboardViewer := False;
  end;
  if FRegisterDragDrop then
  begin
    if HandleAllocated then
      RevokeDragDrop(Handle);
    FRegisterDragDrop := False;
  end;
end;

procedure TCustomShellList.Loaded;
begin
  inherited Loaded;

end;

procedure TCustomShellList.FileNotifyStart;
var
  Path: string;
begin
  Path := GetPIDLPath(FRootPIDL);
  if Path <> '' then
  begin
    FFileNotify := TFileNotifyThread.Create(Path, False, UpdateRoot);
    FFileNotify.Resume;
  end;
end;

procedure TCustomShellList.FileNotifyStop;
begin
  if FFileNotify <> nil then
  begin
    TFileNotifyThread(FFileNotify).Stop;
    FFileNotify := nil;
  end;
end;

procedure TCustomShellList.ClearShellList(IsRefresh: Boolean);

  procedure ClearShellItem(ShellItem: PShellItem);
  begin
    FreePIDL(ShellItem.PIDL);
    Dispose(ShellItem);
  end;

var
  I: Integer;
begin
  if not IsRefresh then
  begin
    FileNotifyStop;
    FreePIDL(FRootPIDL);
    FRootPIDL := nil;
    FShellFolder := nil;
  end;
  for I := 0 to Pred(FShellList.Count) do
    ClearShellItem(FShellList[I]);
end;

var
  GlobalSortDescent: Boolean;
  GlobalSortColumn: TColumnKind;

function ListSortFunc(Item1, Item2: Pointer): Integer;
var
  ShellItem1, ShellItem2: PShellItem;
  DeltaSize: Int64;
  DeltaDateTime: TDateTime;
begin
  if GlobalSortDescent then
  begin
    ShellItem1 := Item2;
    ShellItem2 := Item1;
  end
  else
  begin
    ShellItem1 := Item1;
    ShellItem2 := Item2;
  end;
  Result := 0;
  case GlobalSortColumn of
    ckSize:
      begin
        DeltaSize := ShellItem1.Size - ShellItem2.Size;
        if DeltaSize > 0 then
          Result := 1
        else if DeltaSize < 0 then
          Result := -1;
      end;
    ckType:
      Result :=
        AnsiCompareStr(ShellItem1.TypeName, ShellItem2.TypeName);
    ckModified:
      begin
        DeltaDateTime := ShellItem1.Modified - ShellItem2.Modified;
        if DeltaDateTime > 0 then
          Result := 1
        else if DeltaDateTime < 0 then
          Result := -1;
      end;
  end;
  if Result = 0 then
    Result := AnsiCompareStr(ShellItem1.DisplayName, ShellItem2.DisplayName);
end;

procedure TCustomShellList.Sort;
begin
  GlobalSortDescent := FSortDescent;
  GlobalSortColumn := FSortColumn;
  FShellList.Sort(@ListSortFunc);
end;

procedure TCustomShellList.SaveState;
var
  Item: TListItem;
  ShellItem: PShellItem;
begin
  Item := ItemFocused;
  if Item <> nil then
  begin
    if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
    else
    begin
      ShellItem := FShellList[Item.Index];
      FFocusedPIDL := CopyPIDL(ShellItem.PIDL);
    end;
  end;
  FSelList.Capacity := SelCount;
  Item := GetNextItem(nil, sdAll, [isSelected]);
  while Item <> nil do
  begin
    if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
    else
    begin
      ShellItem := FShellList[Item.Index];
      FSelList.Add(CopyPIDL(ShellItem.PIDL));
    end;
    Item := GetNextItem(Item, sdAll, [isSelected]);
  end;
end;

procedure TCustomShellList.RestoreState(VisibleFocused: Boolean);
var
  FindFocused: Boolean;
  I: Integer;
  ShellItem: PShellItem;
begin
  try
    FindFocused := (FFocusedPIDL <> nil);
    for I := 0 to Pred(FShellList.Count) do
    begin
      ShellItem := FShellList[I];
      if IsPIDLInList(FShellFolder, ShellItem.PIDL, FSelList) then
        Items[I].Selected := True;
      if FindFocused and
        (FShellFolder.CompareIDs(0, ShellItem.PIDL, FFocusedPIDL) = 0) then
      begin
        with Items[I] do
        begin
          Focused := True;
          if VisibleFocused then
            MakeVisible(False);
        end;
        FindFocused := False;
      end;
    end;
  finally
    FreePIDL(FFocusedPIDL);
    FFocusedPIDL := nil;
    for I := 0 to Pred(FSelList.Count) do
      FreePIDL(FSelList[I]);
    FSelList.Clear;
  end;
end;

procedure TCustomShellList.SetRootPIDL(PIDL: PItemIDList);

  function CalcFileSize(FindData: TWin32FindData): Int64;
  begin
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then
      Result := FindData.nFileSizeHigh * MAXDWORD + FindData.nFileSizeLow
    else
      Result := -1;
  end;

  function CalcModifiedDate(FindData: TWin32FindData): TDateTime;
  var
    LocalFileTime: TFileTime;
    Age: Integer;
  begin
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then
    begin
      FileTimeToLocalFileTime(FindData.ftLastWriteTime, LocalFileTime);
      if FileTimeToDosDateTime(LocalFileTime, LongRec(Age).Hi,
        LongRec(Age).Lo) then
      begin
        Result := FileDateToDateTime(Age);
        Exit;
      end;
    end;
    Result := -1;
  end;

  function InFilter(const FileName: string): Boolean;
  var
    I: Integer;
    S: string;
  begin
    Result := False;
    I := LastDelimiter('.\:', FileName);
    if (I > 0) and (FileName[I] = '.') then
    begin
      S := Copy(FileName, I + 1, MaxInt);
      for I := 0 to Pred(FFilter.Count) do
        if AnsiCompareText(FFilter[I], S) = 0 then
        begin
          Result := True;
          Break;
        end;
    end;
  end;

var
  IsRefresh, IsCutPIDL, IsFilter: Boolean;
  RelativePIDL, NewPIDL: PItemIDList;
  EnumIDList: IEnumIDList;
  NumIDs, Attributes, ErrorMode: Cardinal;
  SaveCursor: TCursor;
  ShellItem: PShellItem;
  FindData: TWin32FindData;
begin
  if FUpdating or IsEditing or (FRootPIDL = nil) and (PIDL = nil) then
    Exit;
  HandleNeeded;
  FUpdating := True;
  try
    SaveCursor := Screen.Cursor;
    try
      Screen.Cursor := crHourglass;
      IsRefresh := SamePIDL(FRootPIDL, PIDL);
      try
        if IsRefresh then
          SaveState;
        ClearShellList(IsRefresh);
        FShellList.Clear;
        try
          if not Assigned(PIDL) then
            Exit;
          IsCutPIDL := SamePIDL(FCutPIDL, PIDL);
          IsFilter := FFilter.Count > 0;
          if not IsRefresh then
          begin
            if IsDesktopPIDL(PIDL) then
              FRootPIDL := CreateSpecialPIDL(rsDesktopDirectory)
            else
              FRootPIDL := CopyPIDL(PIDL);
            FShellFolder := GetShellFolder(FRootPIDL);
          end;
          ErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
          try
            if FShellFolder.EnumObjects(Application.Handle, SHCONTF_NONFOLDERS,
              EnumIDList) <> S_OK then
              Exit;
          finally
            SetErrorMode(ErrorMode);
          end;
          NewPIDL := nil;
          while EnumIDList.Next(1, RelativePIDL, NumIDs) = S_OK do
          try
            NewPIDL := ConcatPIDLs(FRootPIDL, RelativePIDL);
            Attributes := GetPIDLAttributes(NewPIDL,
              SFGAO_LINK or SFGAO_FOLDER or SFGAO_FILESYSTEM);
            if not ValueInFlags(SFGAO_LINK, Attributes) and
              not ValueInFlags(SFGAO_FOLDER, Attributes) and
              ValueInFlags(SFGAO_FILESYSTEM, Attributes) then
            begin
              FillChar(FindData, SizeOf(FindData), 0);
              SHGetDataFromIDList(
                FShellFolder,
                RelativePIDL,
                SHGDFIL_FINDDATA,
                @FindData,
                SizeOf(FindData));
              if not IsFilter or InFilter(FindData.cFileName) then
              begin
                ShellItem := New(PShellItem);
                FillChar(ShellItem^, SizeOf(TShellItem), 0);
                ShellItem.PIDL := RelativePIDL;
                RelativePIDL := nil;
                ShellItem.DisplayName := GetPIDLDisplayName(NewPIDL);
                ShellItem.ImageIndex := GetPIDLImage(NewPIDL, False, False);
                ShellItem.Size := CalcFileSize(FindData);
                ShellItem.TypeName := GetPIDLTypeName(NewPIDL);
                ShellItem.Modified := CalcModifiedDate(FindData);
                ShellItem.Cut := IsCutPIDL and InCutList(NewPIDL);
                FShellList.Add(ShellItem);
              end;
            end;
          finally
            FreePIDL(NewPIDL);
            FreePIDL(RelativePIDL);
          end;
          Sort;
        finally
          Items.Clear;
          Items.Count := FShellList.Count;
        end;
      finally
        if IsRefresh then
          RestoreState(True);
      end;
      if (FRootPIDL <> nil) and (FFileNotify = nil) then
        FileNotifyStart;
    finally
      Screen.Cursor := SaveCursor;
    end;
  finally
    FUpdating := False;
  end;
end;

var
  FileSizeFormat: string;

function TCustomShellList.OwnerDataFetch(Item: TListItem;
  Request: TItemRequest): Boolean;

  function FileSizeToStr(Size: Int64): string;
  var
    Dummy: Extended;
  begin
    if Size > 0 then
    begin
      if Size div 1024 = 0 then
        Dummy := 1
      else
        Dummy := Size / 1024;
      Result := Format('%.0n', [Dummy]);
    end
    else
      Result := '0';
  end;

var
  ShellItem: PShellItem;
  I: Integer;
  S: string;
begin
  Result := True;
  if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
    Exit;
  ShellItem := FShellList[Item.Index];
  if irImage in Request then
  begin
    Item.ImageIndex := ShellItem.ImageIndex;
    Item.Cut := ShellItem.Cut;
  end;
  if irText in Request then
    Item.Caption := ShellItem.DisplayName;
  if ViewStyle <> vsReport then
    Exit;
  for I := 1 to Pred(Columns.Count) do
  begin
    S := '';
    case TColumnKind(Columns[I].Tag) of
      ckType:
        S := ShellItem.TypeName;
      ckSize:
        S := Format(FileSizeFormat, [FileSizeToStr(ShellItem.Size)]);
      ckModified:
        S := DateTimeToStr(ShellItem.Modified);
    end;
    Item.SubItems.Add(S);
  end;
end;

function TCustomShellList.OwnerDataFind(Find: TItemFind;
  const FindString: string; const FindPosition: TPoint; FindData: Pointer;
  StartIndex: Integer; Direction: TSearchDirection;
  Wrap: Boolean): Integer;
var
  I: Integer;
  S: string;
  ShellItem: PShellItem;
  Found: Boolean;
begin
  Result := -1;
  if (Find = ifExactString) or (Find = ifPartialString) then
  begin
    I := StartIndex;
    S := AnsiLowerCase(FindString);
    repeat
      if (I >= FShellList.Count) then
        if Wrap and (FShellList.Count > 0) then
          I := 0
        else
          Exit;
      ShellItem := FShellList[I];
      Found := Pos(S, AnsiLowerCase(ShellItem.DisplayName)) = 1;
      Inc(I);
    until Found or (I = StartIndex);
    if Found then
      Result := I - 1;
  end;
end;

procedure TCustomShellList.ColClick(Column: TListColumn);
begin
  if FSortColumn = TColumnKind(Column.Tag) then
    FSortDescent := not FSortDescent
  else
    FSortDescent := False;
  SetSortColumn(TColumnKind(Column.Tag));
end;

procedure TCustomShellList.SetSortColumn(Value: TColumnKind);
begin
  if Value <> FSortColumn then
  begin
    ClearHeader;
    FSortColumn := Value;
  end;
  UpdateHeader;
  if not HandleAllocated then
    Exit;
  FUpdating := True;
  try
    try
      SaveState;
      ListView_SetItemState(Handle, -1, 0, LVIS_SELECTED);
      Sort;
    finally
      RestoreState(False);
    end;
    Invalidate;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellList.ClearHeader;
var
  Item: THDItem;
  I: Integer;
begin
  if FHeaderHandle <> 0 then
    for I := 0 to Pred(Columns.Count) do
      if TColumnKind(Columns[I].Tag) = FSortColumn then
      begin
        FillChar(Item, SizeOf(Item), 0);
        Item.Mask := HDI_FORMAT;
        Header_GetItem(FHeaderHandle, I, Item);
        Item.Mask := HDI_FORMAT;
        Item.fmt := Item.fmt and not (HDF_BITMAP or HDF_BITMAP_ON_RIGHT);
        Header_SetItem(FHeaderHandle, I, Item);
      end;
end;

procedure TCustomShellList.UpdateHeader;
var
  Item: THDItem;
  I: Integer;
begin
  if FHeaderHandle <> 0 then
    for I := 0 to Pred(Columns.Count) do
      if TColumnKind(Columns[I].Tag) = FSortColumn then
      begin
        FillChar(Item, SizeOf(Item), 0);
        Item.Mask := HDI_FORMAT;
        Header_GetItem(FHeaderHandle, I, Item);
        Item.Mask := HDI_FORMAT or HDI_BITMAP;
        Item.fmt := Item.fmt or HDF_BITMAP or HDF_BITMAP_ON_RIGHT;
        Item.hbm := FSortArrows[Ord(FSortDescent)].Handle;
        Header_SetItem(FHeaderHandle, I, Item);
      end;
end;

procedure TCustomShellList.WMParentNotify(var Message: TWMParentNotify);
begin
  inherited;
  with Message do
    if (Event = WM_CREATE) and (FHeaderHandle = 0) then
    begin
      FHeaderHandle := ChildWnd;
      FDefHeaderProc := Pointer(GetWindowLong(FHeaderHandle, GWL_WNDPROC));
      SetWindowLong(FHeaderHandle, GWL_WNDPROC, LongInt(FHeaderInstance));
    end;
end;

procedure TCustomShellList.HeaderWndProc(var Message: TMessage);
begin
  try
    with Message do
    begin
      case Msg of
        WM_NCDESTROY:
          begin
            Result := CallWindowProc(FDefHeaderProc, FHeaderHandle, Msg, WParam, LParam);
            FHeaderHandle := 0;
            FDefHeaderProc := nil;
            Exit;
          end;
      end;
      Result := CallWindowProc(FDefHeaderProc, FHeaderHandle, Msg, WParam, LParam);
    end;
  except
    Application.HandleException(Self);
  end;
end;

procedure TCustomShellList.WMNotify(var Message: TWMNotify);
begin
  inherited;
  if (FHeaderHandle <> 0) and (Message.NMHdr^.hWndFrom = FHeaderHandle) then
    with Message.NMHdr^ do
      case code of
        HDN_ENDTRACK, HDN_ENDDRAG, HDN_DIVIDERDBLCLICK:
          UpdateHeader;
        NM_RCLICK:
          FColumnsContext := True;
      end;
end;

procedure TCustomShellList.DoInfoTip(Item: TListItem; var InfoTip: string);
var
  ShellItem: PShellItem;
  RelativePIDL: PItemIDList;
  QueryInfo: IQueryInfo;
  PInfoTip: PWideChar;
begin
  if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
    Exit;
  ShellItem := FShellList[Item.Index];
  RelativePIDL := ShellItem.PIDL;
  if FShellFolder.GetUIObjectOf(Application.Handle, 1, RelativePIDL,
    IQueryInfo, nil, QueryInfo) = S_OK then
  begin
    if QueryInfo.GetInfoTip(0, PInfoTip) = S_OK then
    begin
      InfoTip := WideString(PInfoTip);
      ShellMalloc.Free(PInfoTip);
    end;
  end;
end;

procedure TCustomShellList.Editing;
var
  Item: TListItem;
begin
  Item := ItemFocused;
  if Item <> nil then
    Item.EditCaption;
end;

function TCustomShellList.GetFilter: string;
begin
  Result := FFilter.CommaText;
end;

procedure TCustomShellList.SetFilter(const Value: string);
begin
  FFilter.CommaText := Value;
  if csDesigning in ComponentState then
    UpdateRoot;
end;

type
  PIDArray = ^TIDArray;
  TIDArray = array of PItemIDList;

function TCustomShellList.GetContextMenu(out ContextMenu: IContextMenu): Boolean;
var
  Count: Integer;
  IDArray: TIDArray;
  Item: TListItem;
  ShellItem: PShellItem;
begin
  ContextMenu := nil;
  Count := SelCount;
  SetLength(IDArray, Count);
  Count := 0;
  Item := GetNextItem(nil, sdAll, [isSelected]);
  while Item <> nil do
  begin
    if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
    else
    begin
      ShellItem := FShellList[Item.Index];
      IDArray[Count] := ShellItem.PIDL;
      Inc(Count);
    end;
    Item := GetNextItem(Item, sdAll, [isSelected]);
  end;
  Result := FShellFolder.GetUIObjectOf(Application.Handle, Count, IDArray[0],
    IID_IContextMenu, nil, Pointer(ContextMenu)) = S_OK;
end;

procedure TCustomShellList.DoVerb(Verb: PChar);
var
  ContextMenu: IContextMenu;
begin
  if FRootPIDL = nil then
    Exit;
  if GetContextMenu(ContextMenu) then
    ContextMenuVerb(ValidParentForm(Self).Handle, ContextMenu, Verb);
end;

procedure TCustomShellList.DoContextPopup(MousePos: TPoint;
  var Handled: Boolean);
var
  ContextMenu: IContextMenu;
  idCmd: Cardinal;
  Verb: string;
begin
  if FColumnsContext then
  begin
    FColumnsContext := False;
    Handled := True;
    with ClientToScreen(MousePos) do
      FColumnsMenu.Popup(x, y);
    Exit;
  end;
  if FRootPIDL = nil then
  begin
    Handled := True;
    Exit;
  end;
  if not FAutoContext or (SelCount = 0) then
  begin
    inherited DoContextPopup(MousePos, Handled);
    Exit;
  end;
  Handled := True;
  if not GetContextMenu(ContextMenu) then
    Exit;
  idCmd := ShowContextMenu(Handle, ContextMenu, FContextMenu2,
    MousePos, CanEdit(nil));
  if idCmd <> 0 then
  begin
    Verb := ContextMenuCommand(ContextMenu, idCmd);
    if Pos('open', LowerCase(Verb)) = 1 then
      ExecuteSelected
    else if SameText(Verb, 'rename') then
      Editing
    else
      ContextMenuVerb(ValidParentForm(Self).Handle,
        ContextMenu, MakeIntResource(idCmd));
  end;
end;

function TCustomShellList.InCutList(PIDL: PItemIdList): Boolean;
var
  I: Integer;
begin
  for I := 0 to Pred(FCutList.Count) do
    if SamePIDL(FCutList[I], PIDL) then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

procedure TCustomShellList.NotifyClipboard;
var
  IsRefresh: Boolean;
  Data: THandle;
  DataPtr: Pointer;
  ChildCount, I: Integer;
  ItemOffset: UINT;
  ParentPIDL, RelativePIDL: PItemIDList;
begin
  IsRefresh := False;
  if FUpdating then
    Exit;
  FUpdating := True;
  try
    if FCutPIDL <> nil then
    begin
      for I := 0 to Pred(FCutList.Count) do
        FreePIDL(FCutList[I]);
      FCutList.Clear;
      IsRefresh := SamePIDL(FRootPIDL, FCutPIDL);
      FreePIDL(FCutPIDL);
      FCutPIDL := nil;
    end;
    if not IsClipboardFormatAvailable(CF_IDLIST) then
      Exit;
    if OpenClipboard(Handle) then
    try
      Data := GetClipboardData(CF_IDLIST);
      if Data = 0 then
        Exit;
      DataPtr := GlobalLock(Data);
      if DataPtr = nil then
        Exit;
      try
        ChildCount := PIDA(DataPtr)^.cidl;
        ItemOffset := UINT(DataPtr) + SizeOf(UINT);
        ParentPIDL := Pointer(UINT(DataPtr) + PUINT(ItemOffset)^);
        for I := 0 to Pred(ChildCount) do
        begin
          Inc(ItemOffset, SizeOf(UINT));
          RelativePIDL :=
            Pointer(UINT(DataPtr) + PUINT(ItemOffset)^);
          FCutList.Add(ConcatPIDLs(ParentPIDL, RelativePIDL));
        end;
        FCutPIDL := CopyPIDL(ParentPIDL);
      finally
        GlobalUnlock(Data);
      end;
    finally
      CloseClipboard;
    end;
    if (FCutPIDL <> nil) and not IsRefresh then
      IsRefresh := SamePIDL(FRootPIDL, FCutPIDL);
  finally
    FUpdating := False;
    if IsRefresh then
      UpdateRoot;
  end;
end;

procedure TCustomShellList.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_MENUSELECT:
      if ShowHint and Assigned(FContextMenu2) then
      begin
        ContextMenuHint(FContextMenu2, TWMMenuSelect(Message));
        Exit;
      end;
    WM_INITMENUPOPUP,
      WM_DRAWITEM,
      WM_MENUCHAR,
      WM_MEASUREITEM:
      if Assigned(FContextMenu2) then
      begin
        FContextMenu2.HandleMenuMsg(Message.Msg, Message.WParam, Message.LParam);
        Message.Result := 0;
        Exit;
      end;
    WM_CHANGECBCHAIN:
      begin
        if THandle(Message.WParam) = FNextViewerHandle then
          FNextViewerHandle := THandle(Message.LParam)
        else if FNextViewerHandle <> 0 then
          with Message do
            SendMessage(FNextViewerHandle, Msg, WParam, LParam);
        Exit;
      end;
    WM_DRAWCLIPBOARD:
      begin
        NotifyClipboard;
        with Message do
          SendMessage(FNextViewerHandle, Msg, WParam, LParam);
        Exit;
      end;
    WM_KEYDOWN, WM_KEYUP:
      if (TWMKey(Message).CharCode in [VK_BACK, VK_INSERT, VK_F5]) and
        not IsEditing and (FShellTree <> nil) then
      begin
        with Message do
          SendMessage(FShellTree.Handle, Msg, WParam, LParam);
        Exit;
      end;
  end;
  inherited WndProc(Message);
end;

function TCustomShellList.CanEdit(Item: TListItem): Boolean;

  function CanRename(Index: Integer): Boolean;
  var
    ShellItem: PShellItem;
    Attributes: Cardinal;
  begin
    Result := False;
    if (Index >= FShellList.Count) or (Index < 0) then
      Exit;
    ShellItem := FShellList[Index];
    Attributes := SFGAO_CANRENAME;
    OleCheck(FShellFolder.GetAttributesOf(1, ShellItem.PIDL, Attributes));
    Result := ValueInFlags(SFGAO_CANRENAME, Attributes);
  end;

begin
  Item := ItemFocused;
  if (Item = nil) or not CanRename(Item.Index) then
  begin
    Result := False;
  end
  else
    Result := inherited CanEdit(Item);
end;

procedure TCustomShellList.Rename(Index: Integer; const NewName: string);
var
  ShellItem: PShellItem;
  NewPIDL, RelativePIDL: PItemIDList;
begin
  if FUpdating or (FRootPIDL = nil) then
    Exit;
  FUpdating := True;
  try
    if (Index >= FShellList.Count) or (Index < 0) then
      Exit;
    ShellItem := FShellList[Index];
    FileNotifyStop;
    NewPIDL := nil;
    if FShellFolder.SetNameOf(
      Application.Handle,
      ShellItem.PIDL,
      POleStr(WideString(NewName)),
      SHGDN_NORMAL,
      RelativePIDL) = S_OK then
    try
      NewPIDL := ConcatPIDLs(FRootPIDL, RelativePIDL);
      FreePIDL(ShellItem.PIDL);
      ShellItem.PIDL := RelativePIDL;
      RelativePIDL := nil;
      ShellItem.DisplayName := GetPIDLDisplayName(NewPIDL);
      ShellItem.ImageIndex := GetPIDLImage(NewPIDL, False, False);
      ShellItem.TypeName := GetPIDLTypeName(NewPIDL);
      UpdateItems(Index, Index);
    finally
      FreePIDL(NewPIDL);
      FreePIDL(RelativePIDL);
    end;
    FileNotifyStart;
  finally
    FUpdating := False;
  end;
end;

procedure TCustomShellList.Edit(const Item: TLVItem);
begin
  with Item do
    if pszText <> nil then
      Rename(iItem, pszText);
end;

procedure TCustomShellList.Execute(Index: Integer);
var
  szPath: array[0..MAX_PATH] of Char;
  SavePath: string;
  ShellItem: PShellItem;
  PIDL: PItemIDList;
begin
  if FRootPIDL = nil then
    Exit;
  GetCurrentDirectory(MAX_PATH, szPath);
  SavePath := szPath;
  try
    SetCurrentDirectory(PChar(GetPIDLPath(FRootPIDL)));
    if (Index >= FShellList.Count) or (Index < 0) then
      Exit;
    ShellItem := FShellList[Index];
    PIDL := ConcatPIDLs(FRootPIDL, ShellItem.PIDL);
    try
      ShellExecutePIDL(PIDL);
    finally
      FreePIDL(PIDL);
    end;
  finally
    SetCurrentDirectory(PChar(SavePath));
  end;
end;

procedure TCustomShellList.ExecuteSelected;
var
  Item: TListItem;
begin
  Item := GetNextItem(nil, sdAll, [isSelected]);
  while Item <> nil do
  begin
    Execute(Item.Index);
    Item := GetNextItem(Item, sdAll, [isSelected]);
  end;
end;

function TCustomShellList.GetItemPath(Index: Integer): string;
var
  PIDL: PItemIDList;
  ShellItem: PShellItem;
begin
  Result := '';
  if FRootPIDL = nil then
    Exit;
  if (Index >= FShellList.Count) or (Index < 0) then
    Exit;
  ShellItem := FShellList[Index];
  PIDL := ConcatPIDLs(FRootPIDL, ShellItem.PIDL);
  try
    Result := GetPIDLPath(PIDL);
  finally
    FreePIDL(PIDL);
  end;
end;

function TCustomShellList.CanPaste: Boolean;
begin
  Result := FCutPIDL <> nil;
end;

{$IFNDEF AS_VCL6}

procedure TCustomShellList.SelectAll;
var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    Items[I].Selected := True;
end;
{$ENDIF}

procedure TCustomShellList.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  case Key of
    VK_DELETE:
      if not IsEditing then
      begin
        DoVerb('delete');
        Key := 0;
      end;
    VK_RETURN:
      if not IsEditing then
      begin
        if ssAlt in Shift then
        else
          ExecuteSelected;
        Key := 0;
      end;
    VK_F2:
      if not IsEditing then
      begin
        Editing;
        Key := 0;
      end;
  end;
end;

procedure TCustomShellList.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case Word(Key) of
    Word(^A):
      if not IsEditing then
      begin
        SelectAll;
        Key := #0;
      end;
    Word(^C):
      if not IsEditing then
      begin
        DoVerb('copy');
        Key := #0;
      end;
    Word(^V):
      if not IsEditing then
      begin
        if FShellTree <> nil then
          FShellTree.DoVerb('paste');
        Key := #0;
      end;
    Word(^X):
      if not IsEditing then
      begin
        DoVerb('cut');
        Key := #0;
      end;
  end;
end;

procedure TCustomShellList.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
end;

procedure TCustomShellList.UpdateRoot;
var
  PIDL: PItemIDList;
  Node: TTreeNode;
begin
  PIDL := nil;
  if (FShellTree <> nil) and FShellTree.HandleAllocated then
  begin
    Node := FShellTree.Items.GetNode(TreeView_GetSelection(FShellTree.Handle));
    if Node <> nil then
      PIDL := Node.Data;
  end;
  SetRootPIDL(PIDL);
end;

procedure TCustomShellList.SetShellTree(Value: TCustomShellTree);
begin
  if FShellTree <> Value then
  begin
    if FShellTree <> nil then
      FShellTree.RemoveFreeNotification(Self);
    FShellTree := Value;
    if FShellTree <> nil then
      FShellTree.FreeNotification(Self);
    if csDesigning in ComponentState then
      UpdateRoot;
  end;
end;

procedure TCustomShellList.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FShellTree then
    begin
      FShellTree := nil;
      UpdateRoot;
    end;
  end;
end;

procedure TCustomShellList.DblClick;
begin
  if Assigned(OnDblClick) then
    OnDblClick(Self)
  else if GetKeyState(VK_MENU) < 0 then
    DoVerb('properties')
  else
    ExecuteSelected;
end;

procedure TCustomShellList.AddColumn(const Name: string; Kind: TColumnKind);
begin
  with Columns.Add do
  begin
    Caption := Name;
    Tag := Ord(Kind);
    if Kind = ckSize then
      Alignment := taRightJustify;
    Width := FColumnsWidth[Kind];
  end;
end;

procedure TCustomShellList.InitColumns;
var
  I: Integer;
  MenuItem: TMenuItem;
begin
  Columns.BeginUpdate;
  try
    for I := 0 to Pred(FColumnsMenu.Items.Count) do
    begin
      MenuItem := FColumnsMenu.Items[I];
      if MenuItem.Checked then
        AddColumn(MenuItem.Caption, TColumnKind(MenuItem.Tag));
    end;
  finally
    Columns.EndUpdate;
  end;
  UpdateHeader;
end;

procedure TCustomShellList.RemoveColumn(Kind: TColumnKind);
var
  I: Integer;
  MenuItem: TMenuItem;
  Column: TListColumn;
begin
  for I := 1 to Pred(FColumnsMenu.Items.Count) do
  begin
    MenuItem := FColumnsMenu.Items[I];
    if TColumnKind(MenuItem.Tag) = Kind then
    begin
      MenuItem.Checked := False;
      Break;
    end;
  end;
  for I := 1 to Pred(Columns.Count) do
  begin
    Column := Columns[I];
    if TColumnKind(Column.Tag) = Kind then
    begin
      FColumnsWidth[Kind] := Column.Width;
      Columns.Delete(I);
      Break;
    end;
  end;
  UpdateHeader;
end;

procedure TCustomShellList.StoreColumnsWidth;
var
  I: Integer;
  Column: TListColumn;
begin
  for I := 0 to Pred(Columns.Count) do
  begin
    Column := Columns[I];
    FColumnsWidth[TColumnKind(Column.Tag)] := Column.Width;
  end;
end;

procedure TCustomShellList.DoColumnMenuClick(Sender: TObject);
var
  MenuItem: TMenuItem;
begin
  MenuItem := TMenuItem(Sender);
  if MenuItem.Checked then
    RemoveColumn(TColumnKind(MenuItem.Tag))
  else
  begin
    MenuItem.Checked := True;
    StoreColumnsWidth;
    Columns.Clear;
    InitColumns;
  end;
end;

function IniReadString(IniFile: TObject; const Section, Ident,
  Default: string): string;
begin
  if IniFile is TRegIniFile then
    Result := TRegIniFile(IniFile).ReadString(Section, Ident, Default)
  else if IniFile is TIniFile then
    Result := TIniFile(IniFile).ReadString(Section, Ident, Default)
  else
    Result := Default;
end;

procedure IniWriteString(IniFile: TObject; const Section, Ident,
  Value: string);
var
  S: string;
begin
  if IniFile is TRegIniFile then
    TRegIniFile(IniFile).WriteString(Section, Ident, Value)
  else
  begin
    S := Value;
    if S <> '' then
    begin
      if ((S[1] = '"') and (S[Length(S)] = '"')) or
      ((S[1] = '''') and (S[Length(S)] = '''')) then
        S := '"' + S + '"';
    end;
    if IniFile is TIniFile then
      TIniFile(IniFile).WriteString(Section, Ident, S);
  end;
end;

function IniReadInteger(IniFile: TObject; const Section, Ident: string;
  Default: Longint): Longint;
begin
  if IniFile is TRegIniFile then
    Result := TRegIniFile(IniFile).ReadInteger(Section, Ident, Default)
  else if IniFile is TIniFile then
    Result := TIniFile(IniFile).ReadInteger(Section, Ident, Default)
  else
    Result := Default;
end;

procedure IniWriteInteger(IniFile: TObject; const Section, Ident: string;
  Value: Longint);
begin
  if IniFile is TRegIniFile then
    TRegIniFile(IniFile).WriteInteger(Section, Ident, Value)
  else if IniFile is TIniFile then
    TIniFile(IniFile).WriteInteger(Section, Ident, Value);
end;

const
  SRegWidth = 'Width';
  SRegOrder = 'Order';
  SRegSort = 'Sort';

const
  IVector: array[Boolean] of Integer = (-1, 1);

procedure TCustomShellList.InternalLoadColumns(IniFile: TObject; const Section: string);
var
  StringList: TStringList;
  I: Integer;
  MenuItem: TMenuItem;
begin
  StringList := TStringList.Create;
  try
    StringList.CommaText := IniReadString(IniFile, Section, SRegWidth, '');
    if StringList.Count = 0 then
      Exit;
    for I := 0 to Pred(Min(StringList.Count, Length(FColumnsWidth))) do
      FColumnsWidth[TColumnKind(I)] := StrToInt(StringList[I]);
    StringList.CommaText := IniReadString(IniFile, Section, SRegOrder, '');
    for I := 0 to Pred(FColumnsMenu.Items.Count) do
    begin
      MenuItem := FColumnsMenu.Items[I];
      MenuItem.Checked := StringList.IndexOf(IntToStr(MenuItem.Tag)) > -1;
    end;
  finally
    StringList.Free;
  end;
  I := IniReadInteger(IniFile, Section, SRegSort, 1);
  FSortDescent := I < 0;
  FSortColumn := TColumnKind(Abs(I) - 1);
  Columns.Clear;
  InitColumns;
  UpdateRoot;
end;

procedure TCustomShellList.LoadColumns(IniFile: TRegIniFile; const Section: string);
begin
  InternalLoadColumns(IniFile, Section);
end;

procedure TCustomShellList.LoadColumns(IniFile: TIniFile; const Section: string);
begin
  InternalLoadColumns(IniFile, Section);
end;

procedure TCustomShellList.InternalSaveColumns(IniFile: TObject; const Section: string);
var
  S: string;
  I: Integer;
  KindLoop: TColumnKind;
begin
  StoreColumnsWidth;
  IniWriteInteger(IniFile, Section, SRegSort,
    (Ord(FSortColumn) + 1) * IVector[FSortDescent = False]);
  S := '';
  for I := 0 to Pred(Columns.Count) do
    S := S + Format('%d,', [Columns[I].Tag]);
  IniWriteString(IniFile, Section, SRegOrder, S);
  S := '';
  for KindLoop := Low(TColumnKind) to High(TColumnKind) do
    S := S + Format('%d,', [FColumnsWidth[KindLoop]]);
  IniWriteString(IniFile, Section, SRegWidth, S);
end;

procedure TCustomShellList.SaveColumns(IniFile: TRegIniFile; const Section: string);
begin
  InternalSaveColumns(IniFile, Section);
end;

procedure TCustomShellList.SaveColumns(IniFile: TIniFile; const Section: string);
begin
  InternalSaveColumns(IniFile, Section);
end;

procedure TCustomShellList.ClearDropTargetItem;
{ ListView_GetNextItem always returns nil for OwnerData = True and
  LVNI_ALL and LVNI_DROPHIGHLITED and Self.DropTarget := nil - does not work }
begin
  ListView_SetItemState(Handle, -1, 0, LVIS_DROPHILITED);
end;

function TCustomShellList.GetDropTarget(MousePos: TPoint): IDropTarget;
var
  Item: TListItem;
  ShellItem: PShellItem;
  PIDL, RelativePIDL: PItemIDList;
  ParentFolder: IShellFolder;
begin
  Result := nil;
  Item := nil;
  try
    if FRootPIDL = nil then
      Exit;
    with ScreenToClient(MousePos) do
      Item := GetItemAt(x, y);
    if Item = nil then
      PIDL := CopyPIDL(FRootPIDL)
    else
    begin
      if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
        Exit;
      ShellItem := FShellList[Item.Index];
      PIDL := ConcatPIDLs(FRootPIDL, ShellItem.PIDL);
    end;
    try
      RelativePIDL := LastPIDL(PIDL);
      ParentFolder := GetParentFolder(PIDL);
      if ParentFolder.GetUIObjectOf(Application.Handle, 1, RelativePIDL,
        IDropTarget, nil, Pointer(Result)) <> S_OK then
        Result := nil;
    finally
      FreePIDL(PIDL);
    end;
  finally
    if (Result <> nil) and (Item <> nil) then
      Self.DropTarget := Item
    else
      ClearDropTargetItem;
  end;
end;

function TCustomShellList.DragEnter(const dataObj: IDataObject;
  grfKeyState: Integer; pt: TPoint; var dwEffect: Integer): HResult;
var
  DropTarget: IDropTarget;
begin
  FDropLeftButton := grfKeyState and MK_LBUTTON <> 0;
  FDataObject := dataObj;
  try
    DropTarget := GetDropTarget(pt);
    if DropTarget = nil then
      Abort;
    Result := DropTarget.DragEnter(dataObj, grfKeyState, pt, dwEffect);
    DropTarget.DragLeave;
    if Result <> S_OK then
      Abort;
  except
    dwEffect := DROPEFFECT_NONE;
    Result := S_OK;
  end;
end;

function TCustomShellList.DragLeave: HResult;
begin
  FDataObject := nil;
  ClearDropTargetItem;
  Result := S_OK;
end;

function TCustomShellList.IDropTarget_DragOver(grfKeyState: Integer;
  pt: TPoint; var dwEffect: Integer): HResult;
var
  DropTarget: IDropTarget;
begin
  try
    DropTarget := GetDropTarget(pt);
    if DropTarget = nil then
      Abort;
    DropTarget.DragEnter(FDataObject, grfKeyState, pt, dwEffect);
    Result := DropTarget.DragOver(grfKeyState, pt, dwEffect);
    DropTarget.DragLeave;
    if Result <> S_OK then
      Abort;
  except
    dwEffect := DROPEFFECT_NONE;
    Result := S_OK;
  end;
end;

function TCustomShellList.Drop(const dataObj: IDataObject;
  grfKeyState: Integer; pt: TPoint; var dwEffect: Integer): HResult;
var
  DropTarget: IDropTarget;
  dwEffect2: Longint;
begin
  try
    try
      DropTarget := GetDropTarget(pt);
      if DropTarget = nil then
        Abort;
      if FDropLeftButton then
        grfKeyState := grfKeyState or MK_LBUTTON;
      dwEffect2 := dwEffect;
      DropTarget.DragEnter(dataObj, grfKeyState, pt, dwEffect2);
      Result := DropTarget.Drop(dataObj, grfKeyState, pt, dwEffect);
      if Result <> S_OK then
        Abort;
    except
      dwEffect := DROPEFFECT_NONE;
      Result := S_OK;
    end;
  finally
    FDataObject := nil;
    ClearDropTargetItem;
  end;
end;

procedure TCustomShellList.OleDragDrop(Index: Integer; IsRightButton: Boolean);
var
  Count: Integer;
  IDArray: TIDArray;
  Item: TListItem;
  ShellItem: PShellItem;
  HR: HResult;
  DataObject: IDataObject;
  DropSource: IDropSource;
  Flags: Cardinal;
  Effect: Longint;
begin
  Count := SelCount;
  SetLength(IDArray, Count);
  Count := 0;
  Item := GetNextItem(nil, sdAll, [isSelected]);
  while Item <> nil do
  begin
    if (Item.Index >= FShellList.Count) or (Item.Index < 0) then
    else
    begin
      ShellItem := FShellList[Item.Index];
      IDArray[Count] := ShellItem.PIDL;
      Inc(Count);
    end;
    Item := GetNextItem(Item, sdAll, [isSelected]);
  end;
  HR := FShellFolder.GetUIObjectOf(Application.Handle, Count, IDArray[0],
    IDataObject, nil, Pointer(DataObject));
  if HR = E_NOINTERFACE then
    Exit;
  OleCheck(HR);
  DropSource := TDropSource.Create(IsRightButton);
  Flags := SFGAO_CANCOPY or SFGAO_CANMOVE or SFGAO_CANLINK;
  OleCheck(FShellFolder.GetAttributesOf(Count, IDArray[0], Flags));
  Effect := 0;
  if ValueInFlags(SFGAO_CANCOPY, Flags) then
    Effect := Effect or DROPEFFECT_COPY;
  if ValueInFlags(SFGAO_CANMOVE, Flags) then
    Effect := Effect or DROPEFFECT_MOVE;
  if ValueInFlags(SFGAO_CANLINK, Flags) then
    Effect := Effect or DROPEFFECT_LINK;
  DoDragDrop(DataObject, DropSource, Effect, Effect);
end;

var
  ShlWAPIModule: THandle;

type
  TSHAutoCompleteFunc = function(hwndEdit: HWND; dwFlags: DWord): LongInt; stdcall;

var
  SHAutoComplete: TSHAutoCompleteFunc;

function EnableAutoComplete(Handle: THandle; FileSystem, URL: Boolean): Boolean;
const
  SHACF_FILESYSTEM = $00000001;
  SHACF_URLHISTORY = $00000002;
  SHACF_URLMRU = $00000004;
  SHACF_USETAB = $00000008;
const
  IFileSystem: array[Boolean] of Cardinal = (0, SHACF_FILESYSTEM);
  IURL: array[Boolean] of Cardinal = (0, SHACF_URLHISTORY or SHACF_URLMRU);
var
  Flags: Cardinal;
begin
  Result := False;
  if Assigned(SHAutoComplete) then
  begin
    Flags := IFileSystem[FileSystem] or IURL[URL];
    if Flags <> 0 then
    begin
      Flags := Flags or SHACF_USETAB;
      Result := SHAutoComplete(Handle, Flags) = 0;
    end;
  end;
end;

initialization
  OleInitialize(nil);
  CF_IDLIST := RegisterClipboardFormat(CFSTR_SHELLIDLIST);

  ShlWAPIModule := GetModule('shlwapi.dll');
  FileSizeFormat := GetStrFromModule(ShlWAPIModule, 260);
  SHAutoComplete := GetProcAddress(ShlWAPIModule, 'SHAutoComplete');

  if Win32MajorVersion < 5 then
    NewFolderName := GetShellString(4177)
  else
    NewFolderName := GetShellString(30320);

finalization
  OleUninitialize;
  if ShlWAPIModule <> 0 then
    FreeLibrary(ShlWAPIModule);

end.

